"""doc
# leanai.data.visualizations.plot_boxes

> Plot detection bounding boxes.
"""
import cv2


def plot_boxes_on_image(image, boxes, color=(0,255,255), thickness=2):
    """
    Plot the boxes onto the image.

    For the boxes a center, size representation is expected: [cx, cy, w, h].

    :param image: The image onto which to draw.
    :param boxes: The boxes which shall be plotted.
    :return: An image with the boxes overlayed over the image.
    """
    for box in boxes:
        start_point = tuple([int(x) for x in box[:2] - box[2:] // 2])
        end_point = tuple([int(x) for x in box[:2] + box[2:] // 2])
        image = cv2.rectangle(image, start_point, end_point, color, thickness)
    return image



import os
import jstyleson

_json_module_library = {}
_native_module_library = {}

def add_module(name=None):
    def wrapper(clazz):
        local_name = clazz.__name__ if name is None else name
        _native_module_library[local_name] = clazz
        return clazz
    return wrapper

def add_lib_from_json(filename):
    def _resolve_includes(data, path):
        if "includes" in data:
            for include in data["includes"]:
                data.update(add_lib_from_json(os.path.join(path, include)))
            del data["includes"]
    with open(filename, 'r') as f:
        data = jstyleson.load(f)
        _json_module_library.update(data)
    path = os.sep.join(filename.split(os.sep)[:-1])
    _resolve_includes(data, path)
    return data



"""doc
# leanai.core.experiment_lightning

> A lightning module that runs an experiment in a managed way.

There is first the Experiment base class from wich all experiments must inherit (directly or indirectly).
"""
import os
import pytorch_lightning as pl
import psutil
import GPUtil
from typing import Any, Tuple
from time import time
from datetime import datetime
from torch.functional import Tensor
from torch.nn import Module
from torch.utils.data.dataloader import DataLoader
from torch.utils.data.dataset import IterableDataset
from pytorch_lightning.utilities.cloud_io import load as pl_load
from torchsummary import summary
try:
    from graphviz import Source
    from torchviz import make_dot
    GRAPHVIZ = True
except:
    GRAPHVIZ = False

from leanai.core.definitions import SPLIT_TEST, SPLIT_TRAIN, SPLIT_VAL
from leanai.core.tensorboard import TensorBoardLogger
from leanai.core.logging import warn, info


def _generate_version() -> str:
    return datetime.fromtimestamp(time()).strftime('%Y-%m-%d_%H.%M.%S')


class Experiment(pl.LightningModule):
    def __init__(self, model: Module=None, loss: Module=None, meta_data_logging=True):
        """
        An experiment base class.

        All experiments must inherit from this.
        
        ```python
        from pytorch_mjolnir import Experiment
        class MyExperiment(Experiment):
            def __init__(self, learning_rate=1e-3, batch_size=32):
                super().__init__(
                    model=Model(),
                    loss=Loss(self)
                )
                self.save_hyperparameters()
        ```

        :param model: The model used for the forward.
        :param loss: The loss used for computing the difference between prediction of the model and the targets.
        :param meta_data_logging: If meta information such as FPS and CPU/GPU Usage should be logged. (Default: True)
        """
        super().__init__()
        self.model = model
        self.loss = loss
        self.meta_data_logging = meta_data_logging

    def run_train(self, name: str, gpus: int, nodes: int, version=None, output_path=os.getcwd(), checkpoint=None):
        """
        Run the experiment.

        :param name: The name of the family of experiments you are conducting.
        :param gpus: The number of gpus used for training.
        :param nodes: The number of nodes used for training.
        :param version: The name for the specific run of the experiment in the family (defaults to a timestamp).
        :param output_path: The path where to store the outputs of the experiment (defaults to the current working directory).
        :param checkpoint: The path to the checkpoint that should be resumed (defaults to None).
            In case of None this searches for a checkpoint in {output_path}/{name}/{version}/checkpoints and resumes it.
            Without defining a version this means no checkpoint can be found as there will not exist a  matching folder.
        """
        if version is None:
            version = _generate_version()
        self.output_path = os.path.join(output_path, name, version)
        if checkpoint is None:
            checkpoint = self._find_checkpoint(name, version, output_path)
        elif not checkpoint.startswith("/"):
            checkpoint = os.path.join(self.output_path, checkpoint)
        self.testing = False    
        if hasattr(self, "example_input_array") and self.example_input_array is not None:
            example_input = self.example_input_array
            if isinstance(example_input, Tensor):
                example_input = (example_input,)
            outp = self(*self.transfer_batch_to_device(example_input))
            if GRAPHVIZ:
                graph = make_dot(outp, params=dict(self.named_parameters()))
                Source(graph).render(os.path.join(self.output_path, "train_graph"))
            try:
                summary(self, [tuple(x.shape[1:]) for x in example_input], device=str(self.device))
            except Exception as e:
                warn(f"Failed to create full summary: {e}")
        trainer = pl.Trainer(
            default_root_dir=output_path,
            max_epochs=getattr(self.hparams, "max_epochs", 1000),
            gpus=gpus,
            num_nodes=nodes,
            logger=TensorBoardLogger(
                save_dir=output_path, version=version, name=name,
                log_graph=hasattr(self, "example_input_array")  and self.example_input_array is not None,
                #default_hp_metric=False
            ),
            resume_from_checkpoint=checkpoint,
            accelerator="ddp" if gpus > 1 else None
        )
        return trainer.fit(self)

    def run_test(self, name: str, gpus: int, nodes: int, version=None, output_path=os.getcwd(), checkpoint=None):
        """
        Evaluate the experiment.

        :param name: The name of the family of experiments you are conducting.
        :param gpus: The number of gpus used for training.
        :param nodes: The number of nodes used for training.
        :param version: The name for the specific run of the experiment in the family (defaults to a timestamp).
        :param output_path: The path where to store the outputs of the experiment (defaults to the current working directory).
        :param evaluate_checkpoint: The path to the checkpoint that should be loaded (defaults to None).
        """
        if version is None:
            version = self._find_version(output_path, name)
        self.output_path = os.path.join(output_path, name, version)
        if checkpoint is None:
            checkpoint = self._find_checkpoint(name, version, output_path)
        elif not checkpoint.startswith("/"):
            checkpoint = os.path.join(self.output_path, checkpoint)
        if checkpoint is None or not os.path.exists(checkpoint):
            raise RuntimeError(f"Checkpoint does not exist: {str(checkpoint)}")
        self.testing = True
        trainer = pl.Trainer(
            default_root_dir=output_path,
            max_epochs=getattr(self.hparams, "max_epochs", 1000),
            gpus=gpus,
            num_nodes=nodes,
            logger=TensorBoardLogger(
                save_dir=output_path, version=version, name=name,
                log_graph=hasattr(self, "example_input_array") and self.example_input_array is not None,
                default_hp_metric=False
            ),
            accelerator="ddp" if gpus > 1 else None
        )
        ckpt = pl_load(checkpoint, map_location=lambda storage, loc: storage)
        self.load_state_dict(ckpt['state_dict'])
        return trainer.test(self)

    def _find_checkpoint(self, name, version, output_path):
        resume_checkpoint = None
        checkpoint_folder = os.path.join(output_path, name, version, "checkpoints")
        if os.path.exists(checkpoint_folder):
            checkpoints = sorted(os.listdir(checkpoint_folder))
            if len(checkpoints) > 0:
                resume_checkpoint = os.path.join(checkpoint_folder, checkpoints[-1])
                info(f"Using Checkpoint: {resume_checkpoint}")
        return resume_checkpoint

    def _find_version(self, output_path, name):
        version = None
        results = os.path.join(output_path, name)
        if os.path.exists(results):
            versions = sorted(os.listdir(results))
            if len(versions) > 0:
                version = os.path.join(results, versions[-1])
                info(f"Using Version: {version}")
        return version

    def prepare_dataset(self, split: str) -> None:
        """
        **ABSTRACT:** Prepare the dataset for a given split.
        
        Only called when cache path is set and cache does not exist yet.
        As this is intended for caching.

        :param split: A string indicating the split.
        """
        raise NotImplementedError("Must be implemented by inheriting classes.")

    def load_dataset(self, split: str) -> Any:
        """
        **ABSTRACT:** Load the data for a given split.

        :param split: A string indicating the split.
        :return: A dataset.
        """
        raise NotImplementedError("Must be implemented by inheriting classes.")

    def prepare_data(self):
        # Prepare the data once (no state allowed due to multi-gpu/node setup.)
        if not self.testing:
            cache_path = getattr(self.hparams, "cache_path", None)
            if cache_path is not None and not os.path.exists(cache_path):
                self.prepare_dataset(SPLIT_TRAIN)
                self.prepare_dataset(SPLIT_VAL)
                assert not getattr(self.hparams, "data_only_prepare", False)

    def training_step(self, batch, batch_idx):
        """
        Executes a training step.

        By default this calls the step function.
        :param batch: A batch of training data received from the train loader.
        :param batch_idx: The index of the batch.
        """
        feature, target = batch
        return self.step(feature, target, batch_idx)

    def validation_step(self, batch, batch_idx):
        """
        Executes a validation step.

        By default this calls the step function.
        :param batch: A batch of val data received from the val loader.
        :param batch_idx: The index of the batch.
        """
        feature, target = batch
        return self.step(feature, target, batch_idx)

    def forward(self, *args, **kwargs):
        """
        Proxy to self.model.
        
        Arguments get passed unchanged.
        """
        if self.model is None:
            raise RuntimeError("You must either provide a model to the constructor or set self.model yourself.")
        return self.model(*args, **kwargs)

    def step(self, feature, target, batch_idx):
        """
        Implementation of a supervised training step.

        The output of the model will be directly given to the loss without modification.

        :param feature: A namedtuple from the dataloader that will be given to the forward as ordered parameters.
        :param target: A namedtuple from the dataloader that will be given to the loss.
        :return: The loss.
        """
        if self.loss is None:
            raise RuntimeError("You must either provide a loss to the constructor or set self.loss yourself.")
        prediction = self(*feature)
        loss = self.loss(prediction, target)
        if self.meta_data_logging:
            self.log_fps()
            self.log_resources()
        self.log('loss/total', loss)
        return loss

    def setup(self, stage=None):
        """
        This function is for setting up the training.

        The default implementation calls the load_dataset function and
        stores the result in self.train_data and self.val_data.
        (It is called once per process.)
        """
        if not self.testing:
            self.train_data = self.load_dataset(SPLIT_TRAIN)
            self.val_data = self.load_dataset(SPLIT_VAL)

    def train_dataloader(self):
        """
        Create a training dataloader.

        The default implementation wraps self.train_data in a Dataloader.
        """
        shuffle = True
        if isinstance(self.train_data, IterableDataset):
            shuffle = False
        return DataLoader(self.train_data, batch_size=self.hparams.batch_size, shuffle=shuffle, num_workers=getattr(self.hparams, "num_workers", 0))

    def val_dataloader(self):
        """
        Create a validation dataloader.

        The default implementation wraps self.val_data in a Dataloader.
        """
        shuffle = True
        if isinstance(self.val_data, IterableDataset):
            shuffle = False
        return DataLoader(self.val_data, batch_size=self.hparams.batch_size, shuffle=shuffle, num_workers=getattr(self.hparams, "num_workers", 0))

    def log_resources(self, gpus_separately=False):
        """
        Log the cpu, ram and gpu usage.
        """
        cpu = psutil.cpu_percent()
        ram = psutil.virtual_memory().used / 1000000000
        self.log("sys/SYS_CPU (%)", cpu)
        self.log("sys/SYS_RAM (GB)", ram)
        total_gpu_load = 0
        total_gpu_mem = 0
        for gpu in GPUtil.getGPUs():
            total_gpu_load += gpu.load
            total_gpu_mem += gpu.memoryUsed
            if gpus_separately:
                self.log("sys/GPU_UTIL_{}".format(gpu.id), gpu.load)
                self.log("sys/GPU_MEM_{}".format(gpu.id), gpu.memoryUtil)
        self.log("sys/GPU_UTIL (%)", total_gpu_load)
        self.log("sys/GPU_MEM (GB)", total_gpu_mem / 1000)

    def log_fps(self):
        """
        Log the FPS that is achieved.
        """
        if hasattr(self, "_iter_time"):
            elapsed = time() - self._iter_time
            fps = self.hparams.batch_size / elapsed
            self.log("sys/FPS", fps)
        self._iter_time = time()

    def train(self, mode=True):
        """
        Set the experiment to training mode and val mode.

        This is done automatically. You will not need this usually.
        """
        if self.logger is not None and hasattr(self.logger, "set_mode"):
            if self.testing:
                self.logger.set_mode("test")
            else:
                self.logger.set_mode("train" if mode else "val")
        super().train(mode)



"""doc
# leanai.model.layers.activation

> Compute an activation function.
"""
from functools import partial
import torch
from torch.nn import Module
from leanai.model.module_registry import add_module


@add_module()
class Activation(Module):
    def __init__(self, activation, **kwargs):
        """
        Supports the activation functions.
    
        :param activation: A string specifying the activation function to use. (Only "relu" and None supported yet.)
        :param **kwargs: The parameters passed to the function (e.g. dim in case of softmax).
        """
        super().__init__()
        self.activation = activation
        self.kwargs = kwargs
        if self.activation is None:
            self.activation = self.activation
        elif self.activation == "relu":
            from torch.nn.functional import relu
            self.activation = relu
        elif self.activation == "softmax":
            from torch.nn.functional import softmax
            self.activation = softmax
        elif self.activation == "sigmoid":
            from torch.nn.functional import sigmoid
            self.activation = sigmoid
        elif self.activation == "tanh":
            from torch.nn.functional import tanh
            self.activation = tanh
        elif self.activation == "elu":
            from torch.nn.functional import elu
            self.activation = elu
        else:
            raise NotImplementedError("Activation '{}' not implemented.".format(self.activation))       
        
    def forward(self, features):
        if self.activation is None:
            return features
        else:
            return self.activation(features, **self.kwargs)



"""doc
# leanai.model.layers.image_conversion

> A simple fully connected layer (aka Linear Layer or Dense).
"""
import torch
from torch.nn import Module
from leanai.core.annotations import RunOnlyOnce
from leanai.model.module_registry import add_module


@add_module()
class ImageConversion(Module):
    def __init__(self, standardize, to_channel_first):
        """
        This layer takes care of image conversion for models.

        :param standardize: (bool) If this is enabled the module will normalize a 0-255 uint image into a float32 image which is normalized and standardized, as with the magic values in pytorch. (Makes them compatible with the imagenet models.)
        :param to_channel_first: (bool) It this is enabled, the image will be converted from h,w,c into the c,h,w format which pytorch uses.
        """
        super().__init__()
        self.standardize = standardize
        self.to_channel_first = to_channel_first

    @RunOnlyOnce
    def build(self, image):
        if self.standardize:
            # Standardization values from torchvision.models documentation
            mean = [0.485, 0.456, 0.406]
            std = [0.229, 0.224, 0.225]
            # Create tensors for a 0-255 value range image.
            self.register_buffer("mean", torch.as_tensor([i * 255 for i in mean], dtype=image.dtype, device=image.device))
            self.register_buffer("std", torch.as_tensor([j * 255 for j in std], dtype=image.dtype, device=image.device))

    def forward(self, image):
        image = image.float()
        self.build(image)
        if self.to_channel_first:
            image = image.permute(0, 3, 1, 2)
        if self.standardize:
            image.sub_(self.mean[None, :, None, None]).div_(self.std[None, :, None, None])
        return image



import os
from leanai.model.models.imagenet import ImagenetModel
from leanai.model.module_registry import add_lib_from_json

model_lib_folder = os.sep.join(os.path.abspath(__file__).split(os.sep)[:-1])
add_lib_from_json(os.path.join(model_lib_folder, "fasterrcnn.jsonc"))
add_lib_from_json(os.path.join(model_lib_folder, "mnist_cnn.jsonc"))



"""doc
# leanai.core.tensorboard

> A logger for tensorboard which is used by the Experiment.
"""
import os
from torch.utils.tensorboard import SummaryWriter

from pytorch_lightning.loggers.base import rank_zero_experiment
from pytorch_lightning.utilities import rank_zero_only
from pytorch_lightning.loggers.tensorboard import TensorBoardLogger as _TensorBoardLogger


class TensorBoardLogger(_TensorBoardLogger):
    """
    The tensorboard logger used in the run_experiment function.

    Normaly you will not need to instantiate this yourself.

    (see documentation of pytorch_lightning.loggers.tensorboard.TensorBoardLogger)
    """
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._mode = "train"
        self._experiment = {}

    @property
    @rank_zero_experiment
    def experiment(self) -> SummaryWriter:
        r"""
        Actual tensorboard object. To use TensorBoard features in your
        :class:`~pytorch_lightning.core.lightning.LightningModule` do the following.

        Example::

            self.logger.experiment.some_tensorboard_function()

        """
        if self._mode in self._experiment:
            return self._experiment[self._mode]

        assert rank_zero_only.rank == 0, 'tried to init log dirs in non global_rank=0'
        if self.root_dir:
            self._fs.makedirs(self.root_dir, exist_ok=True)
        self._experiment[self._mode] = SummaryWriter(log_dir=os.path.join(self.log_dir, self._mode), **self._kwargs)
        return self._experiment[self._mode]

    @rank_zero_only
    def finalize(self, status: str) -> None:
        for experiment in self._experiment.values():
            experiment.flush()
            experiment.close()
        self.save()

    def set_mode(self, mode="train"):
        """
        Set the mode of the logger.

        Creates a subfolder "/{mode}" for tensorboard.
        """
        self._mode = mode



"""doc
# leanai.data.serializer

> The interface for serializing samples.

This is only an interface as a common standard and no implementation, as this varies per problem.
"""
from typing import Any


class ISerializer:
    """
    Interface for a serializer.

    A serializer must implement the __call__ method.
    ```python
    def __call__(self, sample: Any) -> None:
    ````
    The call method gets a sample and has to write it to disk.
    The parameters for writing to disk should be specified in the constructor.
    The sample should contain all info required to write the correct filename (e.g. sample_token).
    """
    def __call__(self, sample: Any) -> None:
        raise NotImplementedError("Must be implemented by subclass!")



"""doc
# leanai.training.losses.sumloss

> An implementation of collector losses like sum, weighted sum, etc.
"""
from typing import Dict
from leanai.training.losses.loss import Loss


class SumLoss(Loss):
    def __init__(self, parent, **losses):
        """
        Compute the sum on the given losses.

        :param **losses: Provide the losses you want to have fused as named parameters to the constructor. Losses get applied to y_pred and y_true, then logged to tensorboard and finally fused.
        """
        super().__init__(parent)
        self.losses = losses
        for k, v in losses.items():
            self.add_module(k, v)

    def forward(self, y_pred, y_true):
        """
        Compute the multiloss using the provided losses.
        
        :param y_pred: The predictions of the network.
        :param y_true: The desired outputs of the network (labels).
        """
        total_loss = 0
        for name, loss in self.losses.items():
            loss_val = loss(y_pred, y_true)
            self.log("loss/{}".format(name), loss_val)
            total_loss += loss_val
        return total_loss


class WeightedSumLoss(Loss):
    def __init__(self, parent, weights: Dict[str, float], **losses):
        """
        Compute the weighted sum on the given losses.

        :param weights: The weights for the losses (the keys must match the keys of **losses).
        :param **losses: Provide the losses you want to have fused as named parameters to the constructor. Losses get applied to y_pred and y_true, then logged to tensorboard and finally fused.
        """
        super().__init__(parent)
        self.losses = losses
        self.weights = weights
        for k, v in losses.items():
            self.add_module(k, v)

    def forward(self, y_pred, y_true):
        """
        Compute the multiloss using the provided losses.
        
        :param y_pred: The predictions of the network.
        :param y_true: The desired outputs of the network (labels).
        """
        total_loss = 0
        for name, loss in self.losses.items():
            loss_val = loss(y_pred, y_true)
            self.log("loss/{}".format(name), loss_val)
            total_loss += self.weights[name] * loss_val
        return total_loss



"""doc
# leanai.model.layers.reshape

> Reshape a tensor.
"""
from torch.nn import Module
from leanai.model.module_registry import add_module


@add_module()
class Reshape(Module):
    def __init__(self, output_shape):
        """
        Reshape a tensor.
    
        A tensor of shape (B, ?) where B is the batch size gets reshaped into (B, output_shape[0], output_shape[1], ...) where the batch size is kept and all other dimensions are depending on output_shape.

        :param output_shape: The shape that the tensor should have after reshaping is (batch_size,) + output_shape (meaning batch size is automatically kept).
        """
        super().__init__()
        self.output_shape = list(output_shape)
        self.output_shape = list(self.output_shape)
        
    def forward(self, features):
        shape = [features.shape[0]] + self.output_shape
        return features.view(shape)



"""doc
# leanai.training.losses.multiloss

> An implementation of the multiloss.
"""
import torch
from torch.nn import Parameter
from leanai.training.losses.loss import Loss
from leanai.training.losses.sumloss import SumLoss


class NormalizedLoss(Loss):
    def __init__(self, parent, loss, name = None, initial_sigma=1):
        """
        Normalize a loss by learning the variance.

        :param parent: The parent for the loss.
        :param loss: The loss that should be weighted by the variance.
        :param name: The name under which to log the sigmas.
        :param initial_sigma: The initial sigma values.
        """
        super().__init__(parent)
        self._loss = loss
        self.name = name
        self.sigma = Parameter(torch.tensor(initial_sigma, dtype=torch.float32, requires_grad=True), requires_grad=True)

    def forward(self, y_pred, y_true):
        """
        Compute the multiloss using the provided losses.
        
        :param y_pred: The predictions of the network.
        :param y_true: The desired outputs of the network (labels).
        """
        loss = self._loss(y_pred, y_true)
        loss = loss / (2 * self.sigma**2)
        if self.name is not None:
            self.log(f"vars/{self.name}_sigma", self.sigma)
            self.log(f"loss/{self.name}_normalized", loss)
        return loss + torch.log(self.sigma**2 + 1)


def MultiLossV2(parent, **losses) -> SumLoss:
    """
    Normalizes the losses by variance estimation and then sums them.

    :param parent: The parent for the loss.
    :param **losses: Provide the losses you want to have fused as named parameters to the constructor. Losses get applied to y_pred and y_true, then logged to tensorboard and finally fused.
    """
    return SumLoss(parent, **{k: NormalizedLoss(parent, v, name=k) for k, v in losses.items()})



from leanai.training.losses.classification import BinaryCrossEntropyLossFromLogits, SparseCrossEntropyLossFromLogits, SparseCategoricalAccuracy
from leanai.training.losses.masking import NaNMaskedLoss, NegMaskedLoss, MaskedLoss
from leanai.training.losses.regression import SmoothL1Loss
from leanai.training.losses.multiloss import MultiLossV2, NormalizedLoss
from leanai.training.losses.sumloss import SumLoss, WeightedSumLoss
from leanai.training.losses.detection import DetectionLoss
from leanai.training.losses.loss import Loss



"""doc
# leanai.data.parser

> The interface and a reference implementation for parsing samples into a usable format.
"""
from typing import Any, Dict, Tuple
import io
import json
import numpy as np
from .data_promise import DataPromise


class IParser:
    """
    Interface for a parser.

    A parser must implement the __call__ method.
    ```python
    def __call__(self, sample: Dict[str, DataPromise]) -> Any:
    ````
    The call method gets the file promises loaded by the file provider and parses them into usable formats.
    """
    def __call__(self, sample: Dict[str, DataPromise]) -> Any:
        raise NotImplementedError("Must be implemented by subclass!")


class Parser(IParser):
    def __init__(self, InputType, OutputType) -> None:
        """
        Parser decodes the samples required to populate the input and output type automatically.

        The Parser should never be instantiated directly.
        You should inherit it and then instantiate the inherited class.

        The Parser automatically only parses those examples which are used in the InputType and OutputType.
        Data which is not used in any of the two will not be parsed thus speeding up the parsing process.

        To achieve this, the parser tries to call functions called "parse_{attribute_name}" (and in case that fails "get_{attribute_name}"),
        where "attribute_name" is the field name in the named tuple.

        For example providing this named tuple:
        `InputType = NamedTuple("InputType", image=np.ndarray)`
        will result in a call to `parse_image` or an error if that function does not exist.
        It will be assumed, that `parse_image` returns an `np.ndarray`.
        If that is not the case, your code may break somewhere.

        **Arguments**
        :param InputType: A definition of a named tuple that defines the input of the neural network.
        :param OutputType: A definition of a named tuple that defines the output of the neural network.
        """
        self.dataset_input_type = InputType
        self.dataset_output_type = OutputType

    def __call__(self, sample: Dict[str, DataPromise]) -> Tuple[Any, Any]:
        dataset_input = self._fill_type_using_parsers(self.dataset_input_type, sample)
        dataset_output = self._fill_type_using_parsers(self.dataset_output_type, sample)
        return dataset_input, dataset_output

    def _fill_type_using_parsers(self, namedtuple_type, sample):
        data = {}
        for k in namedtuple_type._fields:
            parser = getattr(self, "parse_{}".format(k), None)
            if parser is None:
                parser = getattr(self, "get_{}".format(k), None)
            if parser is not None:
                data[k] = parser(sample)
            else:
                raise RuntimeError("Missing parser (parse_{}) for dataset_input_type field: {}".format(k, k))
        return namedtuple_type(**data)

    @staticmethod
    def decode_image(data: bytes, mode: str = "RGB") -> np.ndarray:
        """
        Static method, that allows for decoding an image from bytes in the byte dict.

        This function will be used in parsers, e.g.:
        ```python
        def parse_image(self, sample) -> np.ndarray:
            return BaseParser.decode_image(sample["image"])
        ```

        :param data: Bytes that represent an image. The bytes can be received from a sample in a parse_X function.
        :param mode: The mode in which the image should be laoded. (Default: "RGB")
        :return: An image as a np.ndarray.
        """
        import PIL.Image
        with io.BytesIO(data) as stream:
            img = PIL.Image.open(stream)
            img.load()
            img = img.convert(mode.upper())
        return np.asarray(img)

    @staticmethod
    def decode_numpy(data: bytes, dtype=np.float32) -> np.ndarray:
        """
        Static method, that allows for decoding a numpy file from bytes in the byte dict.

        This function will be used in parsers, e.g.:
        ```python
        def parse_scan(self, sample) -> np.ndarray:
            return BaseParser.decode_numpy(sample["scan"])
        ```

        :param data: Bytes that represent an npy file. The bytes can be received from a sample in a parse_X function.
        :param dtype: The dtype the data has. (Default: np.float32)
        :return: A numpy array.
        """
        return np.frombuffer(data, dtype=dtype)

    @staticmethod
    def decode_text(data: bytes, encoding="utf-8") -> str:
        """
        Static method, that allows for decoding a txt file from bytes in the byte dict.

        This function will be used in parsers, e.g.:
        ```python
        def parse_text(self, sample) -> str:
            return BaseParser.decode_text(sample["label"])
        ```

        :param data: Bytes that represent a txt file. The bytes can be received from a sample in a parse_X function.
        :param encoding: The encoding of the text. (Default: "utf-8")
        :return: The text of the file.
        """
        return data.decode(encoding)

    @staticmethod
    def decode_json(data: bytes, encoding="utf-8"):
        """
        Static method, that allows for decoding a json file from bytes in the byte dict.

        This function will be used in parsers, e.g.:
        ```python
        def parse_annotation(self, sample):
            return BaseParser.decode_json(sample["annotation"])
        ```

        :param data: Bytes that represent a json file. The bytes can be received from a sample in a parse_X function.
        :param encoding: The encoding of the text. (Default: "utf-8")
        :return: A (dict/list) object representing the content of the json.
        """
        return json.loads(Parser.decode_text(data, encoding))



# Inspired by https://github.com/pytorch/vision/blob/6e10e3f88158f12b7a304d3c2f803d2bbdde0823/torchvision/ops/boxes.py#L136
import numpy as np
import torch

from leanai.model.layers.roi_ops import BoxToRoi


convert_to_corners = BoxToRoi()


def similarity_iou_2d(pred_boxes, true_boxes):
    """
    Return intersection-over-union (Jaccard index) of boxes.
    Both sets of boxes are expected to be in (cx, cy, w, h) format.
    Arguments:
        pred_boxes (Tensor[B, 4, N])
        true_boxes (Tensor[B, 4, M])
    Returns:
        iou (Tensor[N, M]): the NxM matrix containing the pairwise IoU values for every element in boxes1 and boxes2
    """
    def area(boxes):
        return (boxes[:, :, 2] - boxes[:, :, 0]) * (boxes[:, :, 3] - boxes[:, :, 1])

    pred_boxes = convert_to_corners(pred_boxes).transpose(1, 2) # BN4
    true_boxes = convert_to_corners(true_boxes).transpose(1, 2) # BN4
    area1 = area(pred_boxes)  # BN
    area2 = area(true_boxes)  # BM
    lt = torch.max(pred_boxes[:,:, None, :2], true_boxes[:,:, :2])  # BNM2
    rb = torch.min(pred_boxes[:,:, None, 2:], true_boxes[:,:, 2:])  # BNM2
    wh = (rb - lt).clamp(min=0)  # BNM2
    inter = wh[:, :, :, 0] * wh[:, :, :, 1]  # BNM
    iou = inter / (area1[:, :, None] + area2 - inter) # BNM
    return iou



"""doc
# leanai.training.losses.detection

> An implementation of a detection loss.
"""
import torch
import numpy as np
from leanai.core import RunOnlyOnce, logging
from leanai.core.logging import debug
from leanai.model.layers import Gather
from leanai.training.losses.loss import Loss
from leanai.training.losses.classification import SparseCrossEntropyLossFromLogits
from leanai.training.losses.regression import SmoothL1Loss
from leanai.training.losses.iou_box_2d import similarity_iou_2d
from leanai.training.losses.masking import NaNMaskedLoss, NegMaskedLoss


def _keep_n_random_trues(arr, num_keep):
    arr_true = arr[arr==True]
    fg_samples = np.random.choice(arr_true.shape[0], num_keep, replace=False)
    arr_true[:] = False
    arr_true[fg_samples] = True
    arr[arr==True] = arr_true
    return arr

def equal_number_sampler(fg, bg, best_indices):
    num = min(int(fg.sum()), int(bg.sum()))
    fg = _keep_n_random_trues(fg, num)
    bg = _keep_n_random_trues(bg, num)
    return fg, bg


class DetectionLoss(Loss):
    def __init__(
        self, parent,
        anchors="anchors", pred_boxes="boxes", pred_class_ids="class_ids", target_boxes="boxes", target_class_ids="class_ids",
        lower_tresh=0.3, upper_tresh=0.5, fg_bg_sampler=equal_number_sampler, similarity_metric=similarity_iou_2d, channel_last_gt=False,
        delta_preds=False, log_delta_preds=False):
        """
        A detection loss.

        :param anchors: The key of the anchors in the predictions.
        """
        super().__init__(parent)
        self.anchors = anchors
        self.pred_boxes = pred_boxes
        self.pred_class_ids = pred_class_ids
        self.target_boxes = target_boxes
        self.target_class_ids = target_class_ids
        self.class_loss = NegMaskedLoss(SparseCrossEntropyLossFromLogits())
        self.center_loss = NaNMaskedLoss(SmoothL1Loss())
        self.size_loss = NaNMaskedLoss(SmoothL1Loss())
        self.delta_preds = delta_preds
        self.log_delta_preds = log_delta_preds

        self.similarity_metric = similarity_metric
        self.lower_tresh = lower_tresh
        self.upper_tresh = upper_tresh
        self.fg_bg_sampler = fg_bg_sampler
        self.channel_last_gt = channel_last_gt
        self.gather = Gather(axis=-1)

    @RunOnlyOnce
    def build_insert_bg_and_ignore(self, classes, boxes):
        # Classes (Background = 0, Ignore = -1)
        classes_shape = list(classes.shape)
        classes_shape[-1] = 2
        class_insertion = np.zeros(classes_shape, dtype=np.int32)
        class_insertion[:, :, 1] = -1 
        self.class_insertion = torch.from_numpy(class_insertion).to(classes.device)

        # Boxes (Background = nan, Ignore = nan)
        boxes_shape = list(boxes.shape)
        boxes_shape[-1] = 2
        boxes_insertion = np.zeros(boxes_shape, dtype=np.float32)
        boxes_insertion[:, :, :] = np.nan
        self.boxes_insertion = torch.from_numpy(boxes_insertion).to(classes.device)

    def insert_bg_and_ignore(self, classes, boxes):
        self.build_insert_bg_and_ignore(classes, boxes)
        classes = torch.cat([self.class_insertion, classes], dim=-1)
        boxes = torch.cat([self.boxes_insertion, boxes], dim=-1)
        return classes, boxes

    def forward(self, y_pred, y_true):
        """
        Compute the detection loss.
        
        :param y_pred: The predictions of the network.
        :param y_true: The desired outputs of the network (labels).
        """
        y_pred_dict = y_pred._asdict()
        y_true_dict = y_true._asdict()
        target_boxes = y_true_dict[self.target_boxes]
        target_classes = y_true_dict[self.target_class_ids]
        anchor_boxes = y_pred_dict[self.anchors]
        if self.channel_last_gt:
            target_boxes = target_boxes.permute(0, 2, 1).contiguous()
            if len(target_classes.shape) == 3:
                target_classes = target_classes.permute(0, 2, 1).contiguous()
        if len(target_classes.shape) == 2:
            target_classes = target_classes[:, None,:]

        target_classes, target_boxes = self.insert_bg_and_ignore(target_classes, target_boxes)

        assignment = self.compute_assignment(anchor_boxes, target_boxes)
        gt_boxes = self.gather(target_boxes, assignment)
        gt_classes = self.gather(target_classes, assignment)
        pred_boxes = y_pred_dict[self.pred_boxes]
        pred_classes = y_pred_dict[self.pred_class_ids]

        class_loss = self.class_loss(pred_classes, gt_classes)
        pos = pred_boxes[:,:2]
        pos_gt = gt_boxes[:,:2]
        pos_a = anchor_boxes[:, :2]
        size = pred_boxes[:,2:]
        size_gt = gt_boxes[:,2:]
        size_a = anchor_boxes[:, 2:]

        if self.delta_preds:
            center_loss = self.center_loss(pos, pos_gt - pos_a)
            size_loss = self.size_loss(size, size_gt - size_a)
        elif self.log_delta_preds:
            center_loss = self.center_loss(pos, (pos_gt - pos_a) / size_a)
            size_loss = self.size_loss(size, torch.log(size_gt / size_a))
        else:
            center_loss = self.center_loss(pos, pos_gt)
            size_loss = self.size_loss(size, size_gt)

        self.log("loss/class({}-{})".format(self.pred_class_ids, self.target_class_ids), class_loss)
        self.log("loss/center({}-{})".format(self.pred_boxes, self.target_boxes), center_loss)
        self.log("loss/size({}-{})".format(self.pred_boxes, self.target_boxes), size_loss)
        return class_loss + center_loss + size_loss

    def compute_assignment(self, anchors, targets):
        indices = torch.zeros(anchors.shape[0], anchors.shape[2], dtype=torch.int64, device=anchors.device, requires_grad=False)
        indices[:] = 1  # = ignore

        # Compute similarity matrix
        similarities = self.similarity_metric(anchors, targets)
        similarities[similarities.isnan()] = -1

        # Rule 2: Assign best matching gt to anchor (each anchor gets 1 gt)
        best_indices = similarities.argmax(axis=-1)
        scores = []
        for batch in range(len(best_indices)):
            scores.append(similarities[batch][range(best_indices.shape[1]), best_indices[batch]])
        scores = torch.stack(scores,dim=0)
        fg = scores > self.upper_tresh
        ignore = (scores > self.lower_tresh) & (scores <= self.upper_tresh)
        bg = scores <= self.lower_tresh

        # Rule 1: Assign best matching anchor to gt (each gt gets assigned (almost) at least once)
        best_per_gt = similarities.argmax(axis=1)
        for batch in range(len(best_per_gt)):
            for idx in range(2, best_per_gt.shape[1]):  # Start with 2 to omit bg and ignore idx.
                if similarities[batch, best_per_gt[batch, idx], idx] > 0:
                    best_indices[batch, best_per_gt[batch, idx]] = idx
                    fg[batch, best_per_gt[batch, idx]] = True
                    bg[batch, best_per_gt[batch, idx]] = False

        if self.fg_bg_sampler is not None:
            fg, bg = self.fg_bg_sampler(fg, bg, best_indices)
        indices[fg] = best_indices[fg]
        indices[bg] = 0  # = bg

        if logging.DEBUG_VERBOSITY:
            debug("Matches")
            debug("GTs: {}".format(targets.shape[1] - 2))
            debug("Foreground: {}".format(fg.long().sum().numpy()))
            debug("Ignore: {}".format(ignore.long().sum().numpy()))
            debug("Background: {}".format(bg.long().sum().numpy()))
            debug("Actual Ignore: {}".format((indices==1).long().sum().numpy()))

        return indices



"""doc
# leanai.model.layers.dropout

> Dropout operation.
"""
from torch.nn import Module
from torch.nn.modules.dropout import Dropout as _Dropout
from leanai.model.module_registry import add_module


@add_module()
class Dropout(Module):
    def __init__(self, p: float, inplace: bool = False):
        """
        Wraps pytorch dropout.
        """
        super().__init__()
        self.dropout = _Dropout(p, inplace)

    def forward(self, features):
        return self.dropout(features)



"""doc
# leanai.data.datasets.coco

> An implementation of the coco dataset.
"""
import cv2
import json
import numpy as np
import os
from typing import Any, Dict, List, NamedTuple
from leanai.core.definitions import SPLIT_TRAIN, SPLIT_VAL, SPLIT_TEST
from leanai.core.logging import warn
from leanai.data.visualizations.plot_boxes import plot_boxes_on_image
from leanai.data.dataset import SequenceDataset
from leanai.data.file_provider import FileProviderSequence
from leanai.data.parser import Parser
from leanai.data.data_promise import DataPromiseFromFile

COCOInputType = NamedTuple("COCOInputType", image=np.ndarray)
COCOOutputType = NamedTuple("COCOOutputType", class_ids=np.ndarray, boxes=np.ndarray, polygons=List[Any])


class COCODataset(SequenceDataset):
    def __init__(self, split, data_path, DatasetInput=COCOInputType, DatasetOutput=COCOOutputType, data_version="2014", model_categories=[], data_image_size=None, shuffle=True) -> None:
        sample_tokens, annotations = self._get_annotations(split, data_path, data_version, model_categories)
        super().__init__(
            file_provider_sequence=_COCOFileProvider(split, data_path, data_version, shuffle, sample_tokens),
            parser=_COCOParser(annotations, DatasetInput, DatasetOutput, data_image_size),
        )

    def _get_annotations(self, split, data_path, version, model_categories):
        image_folder = os.path.join(data_path, "images", f"{split}{version}")
        with open(os.path.join(data_path, "annotations", f"instances_{split}{version}.json"), "r") as f:
            instances = json.loads(f.read())
        class_id_to_category = {0: 0} # Background
        category_id_to_class_id = {0: 0}  # Background
        if len(model_categories) == 0:
            model_categories.append("background")
            for category in instances["categories"]:
                model_categories.append(category["name"])
        for category in instances["categories"]:
            idx = model_categories.index(category["name"])
            if idx > 0:
                class_id_to_category[idx] = category  # List of {'supercategory': 'indoor', 'id': 88, 'name': 'teddy bear'}
                category_id_to_class_id[category["id"]] = idx
        
        images = {}
        for image in instances["images"]:
            images[image["id"]] = image["file_name"]
        
        annotations = {}
        for anno in instances["annotations"]:
            filename = images[anno["image_id"]]
            if not filename in annotations:
                annotations[filename] = []
            if anno["category_id"] in category_id_to_class_id:
                anno["category_id"] = category_id_to_class_id[anno["category_id"]]
                annotations[filename].append(anno)
                # {'segmentation': [[312.29, 562.89, 402.25, 511.49, 400.96, 425.38, 398.39, 372.69, 388.11, 332.85, 318.71, 325.14,
                # 295.58, 305.86, 269.88, 314.86, 258.31, 337.99, 217.19, 321.29, 182.49, 343.13, 141.37, 348.27, 132.37, 358.55,
                # 159.36, 377.83, 116.95, 421.53, 167.07, 499.92, 232.61, 560.32, 300.72, 571.89]],
                # 'area': 54652.9556, 'iscrowd': 0, 'image_id': 480023, 'bbox': [116.95, 305.86, 285.3, 266.03],
                # 'category_id': 58, 'id': 86}
        #print(instances["annotations"][1])
        images_with_no_anno = []
        sample_tokens = []
        for sample_token in os.listdir(image_folder):
            if sample_token in annotations:
                sample_tokens.append(sample_token)
            else:
                images_with_no_anno.append(sample_token)
        if len(images_with_no_anno) > 0:
            warn("Images with no anno: {} (split={})".format(len(images_with_no_anno), split))
        return sample_tokens, annotations


class _COCOFileProvider(FileProviderSequence):
    def __init__(self, split, data_path, version, shuffle, sample_tokens) -> None:
        super().__init__(shuffle=shuffle)
        self.image_folder = os.path.join(data_path, "images", f"{split}{version}")
        self.sample_tokens = sample_tokens
    
    def __getitem__(self, idx) -> Dict[str, Any]:
        sample_token = self.sample_tokens[idx]
        return {
            "image": DataPromiseFromFile(os.path.join(self.image_folder, sample_token)),
            "sample_token": sample_token
        }
    
    def __len__(self) -> int:
        return len(self.sample_tokens)


class _COCOParser(Parser):
    def __init__(self, annotations, InputType, OutputType, data_image_size=None) -> None:
        super().__init__(InputType, OutputType)
        self.annotations = annotations
        self.data_image_size = data_image_size
        self.cheap_cache_sample_token = ""
        self.cheap_cache_image = None

    def parse_image_raw(self, sample):
        sample_token = sample["sample_token"]
        if self.cheap_cache_sample_token == sample_token:
            return self.cheap_cache_image
        image = np.frombuffer(sample["image"].data, np.uint8)
        image = cv2.imdecode(image, cv2.IMREAD_COLOR)
        self.cheap_cache_sample_token = sample_token
        self.cheap_cache_image = image
        return image

    def parse_image(self, sample):
        image = self.parse_image_raw(sample)
        if self.data_image_size is not None:
            image = cv2.resize(image, self.data_image_size)
        return np.copy(image[:,:,::-1])

    def parse_class_ids(self, sample):
        sample_token = sample["sample_token"]
        annos = self.annotations[sample_token]
        class_ids = []
        for anno in annos:
            class_ids.append(anno["category_id"])
        return np.array(class_ids, dtype=np.int32) # FIXME what is the right type?

    def parse_fg_bg_classes(self, sample):
        sample_token = sample["sample_token"]
        annos = self.annotations[sample_token]
        class_ids = []
        for _ in annos:
            class_ids.append(1)
        return np.array(class_ids, dtype=np.int32) # FIXME what is the right type?

    def parse_boxes(self, sample):
        sample_token = sample["sample_token"]
        annos = self.annotations[sample_token]
        boxes = []
        w_scale = 1.0
        h_scale = 1.0
        if self.data_image_size is not None:
            h, w, c = self.parse_image_raw(sample).shape
            wt, ht = self.data_image_size
            w_scale = wt / w
            h_scale = ht / h
        for anno in annos:
            center = [c + s//2 for c, s in zip(anno["bbox"][:2], anno["bbox"][2:])]
            size = anno["bbox"][2:]
            center[0] *= w_scale
            center[1] *= h_scale
            size[0] *= w_scale
            size[1] *= h_scale
            boxes.append(center + size)
        return np.array(boxes, dtype=np.float32)

    def parse_polygons(self, sample):
        sample_token = sample["sample_token"]
        annos = self.annotations[sample_token]
        masks = []
        for anno in annos:
            # FIXME scaling not supported
            masks.append(anno["segmentation"])
        return masks


def _test_visualization(data_path):
    import matplotlib.pyplot as plt
    dataset = COCODataset(SPLIT_TRAIN, data_path, data_version="2014", data_image_size=(800,600))
    inputs, target = dataset[0]
    plt.title(target.class_ids)
    image = inputs.image
    image = plot_boxes_on_image(image, target.boxes)
    plt.imshow(image)
    plt.show()

if __name__ == "__main__":
    import sys
    data_path = sys.argv[1] if len(sys.argv) > 1 else os.path.join(os.environ["DATA_PATH"], "COCO")
    _test_visualization(data_path)



"""doc
# leanai.core.annotations

> A collection of helpful annotations.
"""
from abc import ABC, abstractmethod


class _ClassDecorator(ABC):
    def __get__(self, obj, objtype):
        """
        A class decorator is a base class that is used for all annotations that should be usable with python classes.
        Regular annotations will not work with classes.

        This is a helper class that can be used when writing annotations.
        """
        import functools
        return functools.partial(self.__call__, obj)

    @abstractmethod
    def __call__(self, *args, **kwargs):
        pass


class RunOnlyOnce(_ClassDecorator):
    def __init__(self, f):
        """
        A decorator that ensures a function in an object gets only called exactly once.

        The run only once annotation is fundamental for the build function pattern, whereas it allows to write a function which is only called once, no matter how often it gets called. This behaviour is very usefull for creating variables on the GPU only once in the build and not on every run of the neural network.
        This is for use with the build function in a module. Ensuring it only gets called once and does not eat memory on the gpu.
        For example, using this on a function which prints the parameter only yields on printout, even though the function gets called multiple times.

        :param f: The function that should be wrapped.
        """
        self.f = f
        self.called = {}

    def __call__(self, *args, **kwargs):
        if args[0] not in self.called:
            res = self.f(*args, **kwargs)
            self.called[args[0]] = res
            return res
        else:
            return self.called[args[0]]



from leanai.model.layers.activation import Activation
from leanai.model.layers.batch_normalization import BatchNormalization
from leanai.model.layers.convolution import Conv1D, Conv2D
from leanai.model.layers.dense import Dense
from leanai.model.layers.detection import DetectionHead, GridAnchorGenerator
from leanai.model.layers.dropout import Dropout
from leanai.model.layers.flatten import Flatten
from leanai.model.layers.image_conversion import ImageConversion
from leanai.model.layers.pooling import GlobalAveragePooling1D, GlobalAveragePooling2D, MaxPooling1D, MaxPooling2D, AdaptiveAvgPool2D
from leanai.model.layers.reshape import Reshape
from leanai.model.layers.roi_ops import RoiAlign, RoiPool, BoxToRoi
from leanai.model.layers.selection import TopKIndices, Gather, GatherTopKIndices
from leanai.model.layers.sequential import Sequential
from leanai.model.layers.tensor_combiners import Concat, Stack



"""doc
# leanai.core.cli

> The command line interface for leanai.

By using this package you will not need to write your own main for most networks. This helps reduce boilerplate code.
"""
import argparse
import os
import leanai.core.logging as logging


def set_seeds():
    """
    Sets the seeds of torch, numpy and random for reproducability.
    """
    import torch
    torch.manual_seed(0)
    import numpy as np
    np.random.seed(0)
    import random
    random.seed(0)


def run(experiment_class):
    """
    You can use this main function to make your experiment runnable with command line arguments.

    Simply add this to the end of your experiment.py file:

    ```python
    if __name__ == "__main__":
        from pytorch_mjolnir import run
        run(MyExperiment)
    ```

    Then you can call your python file from the command line and use the help to figure out the parameters.
    ```bash
    python my_experiment.py --help
    ```
    """
    gpus = 1
    if "SLURM_GPUS" in os.environ:
        gpus = int(os.environ["SLURM_GPUS"])
    nodes = 1
    if "SLURM_NODES" in os.environ:
        nodes = int(os.environ["SLURM_NODES"])
    output_path = "logs"
    if "RESULTS_PATH" in os.environ:
        output_path = os.environ["RESULTS_PATH"]
    
    parser = argparse.ArgumentParser(description='The main entry point for the script.')
    parser.add_argument('--name', type=str, required=True, help='The name for the experiment.')
    parser.add_argument('--version', type=str, required=False, default=None, help='The version that should be used (defaults to timestamp).')
    parser.add_argument('--output', type=str, required=False, default=output_path, help='The name for the experiment (defaults to $RESULTS_PATH or "logs").')
    parser.add_argument('--mode', type=str, required=False, default="train", help='The action that should be executed (in the config/experiment calls the function named like this).')
    parser.add_argument('--gpus', type=int, required=False, default=gpus, help='Number of GPUs that can be used.')
    parser.add_argument('--nodes', type=int, required=False, default=nodes, help='Number of nodes that can be used.')
    parser.add_argument('--checkpoint', type=str, required=False, default=None, help='A specific checkpoint to load. If not provided it tries to load latest if any exists.')
    parser.add_argument('--debug', action='store_true', help='This flag will make leanai print debug messages.')
    parser.add_argument('--device', type=str, default=None, required=False, help='CUDA device id if you want to use a specific one.')
    args, other_args = parser.parse_known_args()
    kwargs = _parse_other_args(other_args)

    if args.debug:
        logging.DEBUG_VERBOSITY = args.debug
        logging.debug(f"Set DEBUG_VERBOSITY={logging.DEBUG_VERBOSITY}")
    
    if args.device:
        os.environ['CUDA_VISIBLE_DEVICES'] = args.device

    for k, v in args._get_kwargs():
        logging.info(f"Arg: --{k} {v}")
    for k, v in kwargs.items():
        logging.info(f"Arg: --{k} {v}")
    _instantiate_and_run(experiment_class, name=args.name, mode=args.mode, version=args.version, output=args.output, checkpoint=args.checkpoint, gpus=args.gpus, nodes=args.nodes, **kwargs)


def _instantiate_and_run(
        experiment_class,
        name: str,
        mode: str = "train",
        version: str = None,
        output: str = "logs",
        checkpoint: str = None,
        gpus: int = 1,
        nodes: int = 1,
        **kwargs
    ):
    set_seeds()
    experiment = experiment_class(**kwargs)
    fun = getattr(experiment, f"run_{mode}", None)
    if fun is None:
        raise NotImplementedError(f"No function run_{fun} in your experiment, but --mode='{fun}' requires that function to exist.")
    else:
        return fun(name=name, version=version, output_path=output, checkpoint=checkpoint, gpus=gpus, nodes=nodes)


def _parse_other_args(other_args):
    kwargs = {}
    for arg in other_args:
        parts = arg.split("=")
        k = parts[0]
        if k.startswith("--"):
            k = k[2:]
        if len(parts) == 1:
            v = True
        else:
            v = "=".join(parts[1:])
            if v.startswith('"') or v.startswith("'"):
                v = v[1:-2]
            elif v == "True":
                v = True
            elif v == "False":
                v = False
            else:
                try:
                    v = int(v)
                except ValueError:
                    try:
                        v = float(v)
                    except ValueError:
                        pass
        kwargs[k] = v
    return kwargs



"""doc
# leanai.core.logging

> Logging for tensorflow and pytorch.

## Outline

This package helps with logging while training and managing checkpoints (creation, eta estimation, etc.).

1. Global Flags
2. Time Helpers
3. Logging
"""
import datetime
import json
import os
import shutil
import time
import fnmatch
from typing import List, Union
from functools import reduce

import time as __time
import datetime as __datetime

import numpy as np
try:
    import matplotlib.pyplot as plt
except:
    pass
import imageio


"""doc
# Global Flags

These flags allow controlling what gets printed when using status, info, warn and error function. Debug verbosity enables a lot of logging leanai internally.

```python
DEBUG_VERBOSITY = False
PRINT_STATUS = True
PRINT_INFO = True
PRINT_WARN = True
PRINT_ERROR = True
```

For logging the code an ignore list is used to find out what files to ignore. If you need to change those files you can modify the following global variable of this module, before setting up the logger.

```python
PYTHON_IGNORE_LIST = ["__pycache__", "*.pyc", ".ipynb_checkpoints", "checkpoints", "logs", "dist", "docs", "*.egg-info", "tfrecords", "*.code-workspace", ".git"]
```

For example:

```python
from leanai.core import logging
logging.PYTHON_IGNORE_LIST.append("*.my_extension")
logging.set_logger("...")
```
"""
__logfile = None
__log_buffer = []
__last_progress = 0
__last_update = time.time()


DEBUG_VERBOSITY = False
PRINT_STATUS = True
PRINT_INFO = True
PRINT_WARN = True
PRINT_ERROR = True

PYTHON_IGNORE_LIST = ["__pycache__", "*.pyc", ".ipynb_checkpoints", "checkpoints", "logs", "dist", "docs", "*.egg-info",
                      "tfrecords", "*.code-workspace", ".git"]

"""doc
# Time Helpers

These functions help handling time formating and computations.
"""
def format_time(t: float) -> str:
    """
    Format a time duration in a readable format.
    
    :param t: The duration in seconds.
    :return: A human readable string.
    """
    hours, remainder = divmod(t, 3600)
    minutes, seconds = divmod(remainder, 60)
    return '%d:%02d:%02d' % (hours, minutes, seconds)


def get_timestamp() -> str:
    """
    Create a string for the current timestamp.
    
    :return: Current date and time as a string suitable for a logfolder filename.
    """
    time_stamp = __datetime.datetime.fromtimestamp(__time.time()).strftime('%Y-%m-%d_%H.%M.%S')
    return time_stamp


def compute_eta(progress, last_progress, last_update):
    """
    Compute the ETA.
    
    Uses the current timestamp the last_update and the difference between progresses to estimate the remaining time.
    
    :param progress: The current progress (between 0 and 1).
    :param last_progress: The progress measured at the last step (between 0 and 1).
    :param last_update: The timestamp when the last progress was measured.
    """
    assert 0 <= progress <= 1, progress
    assert 0 <= last_progress <= 1, last_progress
    
    delta_t = time.time() - last_update
    delta_p = max(progress - last_progress, 1E-6)
    eta = (1.0 - progress) / delta_p * delta_t
    return eta


"""doc
# Logging
"""
def __get_all_files(root: str = None, forbidden_list: List[str] = PYTHON_IGNORE_LIST) -> List[str]:
    if root is None:
        root = os.getcwd()
    
    def _is_not_in_forbidden_list(candidate: str) -> bool:
        candidate = candidate.replace("\\", "/")
        res = map(lambda x: not fnmatch.fnmatch(candidate, x), forbidden_list)
        return reduce(lambda x, y: (x and y), res)

    def _filter_files(files):
        return list(filter(_is_not_in_forbidden_list, files))
    
    def _filter_subdirs(subdirs):
        filtered = filter(_is_not_in_forbidden_list, subdirs)
        return list(filter(lambda x: not x.startswith("."), filtered))
    
    def _join_filenames(root, path, files):
        root_with_sep = root + os.sep
        files = map(lambda name: os.path.join(path, name).replace(root_with_sep, ""), files)
        return list(files)
    
    all_files = []
    for path, subdirs, files in os.walk(root):
        files = _filter_files(files)
        subdirs[:] = _filter_subdirs(subdirs)
        all_files.extend(_join_filenames(root, path, files))
    return all_files


def _log_code(output_dir: str, overwrite_existing=False) -> None:
    """
    Log the code of the current working directory into output directory.

    :param output_dir: The directory where to copy all code.
    :param overwrite_existing: When set to true it overwrites existing code copies.
    """
    def _get_backup_path(fname: str) -> str:
        return os.path.join(os.path.normpath(output_dir), fname)

    def _create_missing_dir(fname: str) -> None:
        path = os.path.dirname(fname)
        if not os.path.exists(path):
            os.makedirs(path)

    def _backup(fname: str) -> None:
        target = _get_backup_path(fname)
        _create_missing_dir(target)
        shutil.copyfile(fname, target)

    if overwrite_existing or not os.path.exists(output_dir):
        for f in __get_all_files():
            _backup(f)


def set_logger(log_file: str) -> None:
    """
    Setup the logger.

    Creates the log folder, a src folder inside the log folder where it copies the current working directory.

    :param log_file: File where the warns, errors, etc. should be written. The folder in which the log file is is used for checkpointing.
    """
    def _set_logfile(log_file):
        global __logfile
        if __logfile is not None:
            raise RuntimeError("You must not setup logger twice!")
        __logfile = log_file

    def _create_log_folder():
        log_folder = get_log_path()
        assert log_folder is not None
        if not os.path.exists(log_folder):
            os.makedirs(log_folder)
    
    def _flush_log_buffer():
        assert __logfile is not None
        with open(__logfile, "a") as f:
            for data in __log_buffer:
                f.write(data + "\n")

    _set_logfile(log_file)
    _create_log_folder()
    _flush_log_buffer()
    log_folder = get_log_path()
    assert log_folder is not None
    _log_code(os.path.join(log_folder, "src_{}".format(get_timestamp())))
    log_progress(goal="waiting", progress=0, score=0)


def close(reason: str = None) -> None:
    """
    Close the logger for a given reason.
    
    If none is provided there is no final progress written. Provide a reason, if you do not manually set the final progress before.
    A training loop typically manually sets the progress, so you will not need a reason in that case.
    
    :param reason: The reason for the closing of the logger. It is recommended to use "done", "paused", "failed" as reason.
    """
    global __logfile
    global __log_buffer
    global __last_progress
    global __last_update
    if __logfile is None:
        raise RuntimeError("You must setup the logger before you can close it!")
    if reason is not None:
        log_progress(goal=reason, progress=1.0, score=0.0)

    __log_buffer = []
    __logfile = None
    __last_progress = 0
    __last_update = time.time()


def get_log_path() -> Union[str, None]:
    """
    Gets the log path based on the logfile.
    
    :return: The path containing the logfile.
    """
    if __logfile is None:
        return None
    return os.path.dirname(__logfile)


def create_checkpoint_structure() -> None:
    """
    Create a checkpoint structure in the log folder.
    
    * train: Folder for train split tensorboard logs.
    * val: Folder for val split tensorboard logs.
    * test: Folder for test split tensorboard logs.
    * checkpoints: Folder for the leanai, pytorch or tensorboard checkpoints.
    * images: Folder for images that are generated by your code. (use log_image)
    """
    logfolder = get_log_path()
    assert logfolder is not None
    if not os.path.exists(os.path.join(logfolder, "train")):
        os.makedirs(os.path.join(logfolder, "train"))
    if not os.path.exists(os.path.join(logfolder, "val")):
        os.makedirs(os.path.join(logfolder, "val"))
    if not os.path.exists(os.path.join(logfolder, "test")):
        os.makedirs(os.path.join(logfolder, "test"))
    if not os.path.exists(os.path.join(logfolder, "checkpoints")):
        os.makedirs(os.path.join(logfolder, "checkpoints"))
    if not os.path.exists(os.path.join(logfolder, "images")):
        os.makedirs(os.path.join(logfolder, "images"))


def log_progress(goal: str, progress: float, score: Union[float, np.number]) -> None:
    """
    Update the progress value. Automatically also computes the ETA and updates it in the logs.
    
    :param goal: The current goal it tries to reach. ("training", "validating", "pause", "waiting", "done")
    :param progress: A value between 0 and 1 indicating the progress, where 1 means done. The value should grow monotonic.
    :param score: The score that the approach achieved.
    """
    
    def _get_ailab_logfile() -> Union[str, None]:
        return os.path.join(os.path.dirname(__logfile), "ailab.log")
    
    def _write_log(log_entry: object) -> None:
        out_str = json.dumps(log_entry)
        logfile = _get_ailab_logfile()
        assert logfile is not None
        with open(logfile, "a") as f:
            f.write(out_str + "\n")
            
    def _get_eta():
        global __last_progress
        global __last_update
        if goal == "pause" or goal == "waiting":
            return 0
        eta = compute_eta(progress, __last_progress, __last_update)
        __last_update = time.time()
        __last_progress = progress
        return eta

    if __logfile is None:
        warn("You should setup logger before using it. Call leanai.core.logger.set_logger(...).")
    else:
        log_entry = {
            "timestamp": "{}".format(datetime.datetime.now()),
            "eta": format_time(_get_eta()),
            "goal": str(goal),
            "score": int(score * 1000) / 1000,
            "progress": int(progress * 1000) / 1000
        }
        _write_log(log_entry)


def log_image(*, name: str, data: np.ndarray = None) -> None:
    """
    Log an image.
    
    :param name: The name of the image (e.g. "example.png").
    :param data: The data (optional) if none is provided it is assumed that a pyplot figure should be saved.
    """
    logfolder = get_log_path()
    if logfolder is None:
        warn("Cannot log images when logger is not setup. Call logger.setup first")
        return
    fname = os.path.join(logfolder, "images", name)
    image_folder = os.path.dirname(fname)
    if not os.path.exists(image_folder):
        os.makedirs(image_folder)
    if data is None:
        plt.savefig(fname)
    else:
        imageio.imwrite(fname, data)


def status(msg: str, end: str = "\n") -> None:
    """
    Print something with a timestamp.
    Useful for logging.
    Leanai internally uses this for all its log messages.

    :param msg: The message to print.
    :param end: The line ending. Defaults to "\n" but can be set to "" to not have a linebreak.
    """
    if PRINT_STATUS:
        time_stamp = __datetime.datetime.fromtimestamp(__time.time()).strftime('%Y-%m-%d %H:%M:%S')
        data = "[{}] STAT {}".format(time_stamp, msg)
        print("\r{}".format(data), end=end)
        if end != "":
            if __logfile is not None:
                with open(__logfile, "a") as f:
                    f.write(data + "\n")
            else:
                __log_buffer.append(data)


def info(msg: str, end: str= "\n") -> None:
    """
    Print something with a timestamp.
    Useful for logging.
    Leanai internally uses this for all its log messages.

    :param msg: The message to print.
    :param end: The line ending. Defaults to "\n" but can be set to "" to not have a linebreak.
    """
    if PRINT_INFO:
        time_stamp = __datetime.datetime.fromtimestamp(__time.time()).strftime('%Y-%m-%d %H:%M:%S')
        data = "[{}] INFO {}".format(time_stamp, msg)
        print("\r{}".format(data), end=end)
        if __logfile is not None:
            with open(__logfile, "a") as f:
                f.write(data + "\n")
        else:
            __log_buffer.append(data)


def warn(msg: str, end: str = "\n") -> None:
    """
    Print something with a timestamp.
    Useful for logging.
    Leanai internally uses this for all its log messages.

    :param msg: The message to print.
    :param end: The line ending. Defaults to "\n" but can be set to "" to not have a linebreak.
    """
    if PRINT_WARN:
        time_stamp = __datetime.datetime.fromtimestamp(__time.time()).strftime('%Y-%m-%d %H:%M:%S')
        data = "[{}] WARN {}".format(time_stamp, msg)
        print("\r{}".format(data), end=end)
        if __logfile is not None:
            with open(__logfile, "a") as f:
                f.write(data + "\n")
        else:
            __log_buffer.append(data)


def debug(msg: str, end: str = "\n") -> None:
    """
    Print something with a timestamp.
    Useful for logging.
    Leanai internally uses this for all its log messages.

    :param msg: The message to print.
    :param end: The line ending. Defaults to "\n" but can be set to "" to not have a linebreak.
    """
    if DEBUG_VERBOSITY:
        time_stamp = __datetime.datetime.fromtimestamp(__time.time()).strftime('%Y-%m-%d %H:%M:%S')
        data = "[{}] DEBUG {}".format(time_stamp, msg)
        print("\r{}".format(data), end=end)
        if __logfile is not None:
            with open(__logfile, "a") as f:
                f.write(data + "\n")
        else:
            __log_buffer.append(data)


def error(msg: str, end: str = "\n") -> None:
    """
    Print something with a timestamp.
    Useful for logging.
    Leanai internally uses this for all its log messages.

    :param msg: The message to print.
    :param end: The line ending. Defaults to "\n" but can be set to "" to not have a linebreak.
    """
    if PRINT_ERROR:
        time_stamp = __datetime.datetime.fromtimestamp(__time.time()).strftime('%Y-%m-%d %H:%M:%S')
        data = "[{}] ERROR {}".format(time_stamp, msg)
        print("\r{}".format(data), end=end)
        if __logfile is not None:
            with open(__logfile, "a") as f:
                f.write(data + "\n")
        else:
            __log_buffer.append(data)



"""doc
# leanai.data.data_promise

> The promise object for data can be used to abstract how the data is loaded and only load it lazy.
"""


class DataPromise:
    @property
    def data(self) -> bytes:
        """
        Get the data as raw bytes.

        Must be implemented by all subclasses.
        """
        raise NotImplementedError("The data property must be implemented by subclasses.")


class DataPromiseFromFile(DataPromise):
    def __init__(self, filename) -> None:
        """
        A promise on the data in a file.
        Only loads the data on access and buffers it then.
        """
        self._filename = filename
        self._buffer = None

    @property
    def data(self) -> bytes:
        if self._buffer is None:
            with open(self._filename, "rb") as f:
                self._buffer = f.read()
        return self._buffer


class DataPromiseFromBytes(DataPromise):
    def __init__(self, bytes) -> None:
        """
        A promise on bytes that are already loaded.

        Can be usefull when promising data from a tar stream.
        """
        self._bytes = bytes

    @property
    def data(self) -> bytes:
        return self._bytes



"""doc
# leanai.training.losses.classification

> All losses related to classification problems.
"""
from torch.nn import Module
from torch import Tensor
from torch.nn import CrossEntropyLoss
from torch.nn import BCEWithLogitsLoss


class SparseCrossEntropyLossFromLogits(Module):
    def __init__(self, model=None, reduction: str = "mean"):
        """
        Compute a sparse cross entropy.
        
        This means that the preds are logits and the targets are not one hot encoded.
        
        :param reduction: Specifies the reduction to apply to the output: `'none'` | `'mean'` | `'sum'`. `'none'`: no reduction will be applied, `'mean'`: the sum of the output will be divided by the number of elements in the output, `'sum'`: the output will be summed. Default: 'mean'.
        """
        super().__init__()
        self.loss_fun = CrossEntropyLoss(reduction=reduction)
        
    def forward(self, y_pred, y_true):
        """
        Compute the sparse cross entropy assuming y_pred to be logits.
        
        :param y_pred: The predictions of the network. Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        :param y_true: The desired outputs of the network (labels). Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        """
        if not isinstance(y_true, Tensor):
            y_true = y_true.class_id
        if not isinstance(y_pred, Tensor):
            y_pred = y_pred.class_id
        y_true = y_true.long()
        if len(y_true.shape) == len(y_pred.shape) and y_true.shape[1] == 1:
            y_true = y_true[:, 0]
        return self.loss_fun(y_pred, y_true)


class BinaryCrossEntropyLossFromLogits(Module):
    def __init__(self, model=None, reduction: str = "mean"):
        """
        Compute a binary cross entropy.
        
        This means that the preds are logits and the targets are a binary (1 or 0) tensor of same shape as logits.

        :param reduction: Specifies the reduction to apply to the output: `'none'` | `'mean'` | `'sum'`. `'none'`: no reduction will be applied, `'mean'`: the sum of the output will be divided by the number of elements in the output, `'sum'`: the output will be summed. Default: 'mean'.
        """
        super().__init__()
        self.loss_fun = BCEWithLogitsLoss(reduction=reduction)

    def forward(self, y_pred, y_true):
        """
        Compute the sparse cross entropy assuming y_pred to be logits.
        
        :param y_pred: The predictions of the network. Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        :param y_true: The desired outputs of the network (labels). Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        """
        if not isinstance(y_true, Tensor):
            y_true = y_true.class_id
        if not isinstance(y_pred, Tensor):
            y_pred = y_pred.class_id
        return self.loss_fun(y_pred, y_true)


class SparseCategoricalAccuracy(Module):
    def __init__(self, model=None, reduction: str = "mean", axis=-1):
        """
        Compute the sparse mean squared error.
        
        Sparse means that the targets are not one hot encoded.
        
        :param reduction: Specifies the reduction to apply to the output: `'none'` | `'mean'` | `'sum'`. `'none'`: no reduction will be applied, `'mean'`: the sum of the output will be divided by the number of elements in the output, `'sum'`: the output will be summed. Default: 'mean'.
        """
        super().__init__()
        self.reduction = reduction
        self.axis = axis

    def forward(self, y_pred, y_true):
        """
        Compute the sparse categorical accuracy.
        
        :param y_pred: The predictions of the network. Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        :param y_true: The desired outputs of the network (labels). Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        :param axis: (Optional) The axis along which to compute the sparse categorical accuracy.
        """
        if not isinstance(y_true, Tensor):
            y_true = y_true.class_id
        if not isinstance(y_pred, Tensor):
            y_pred = y_pred.class_id

        pred_class = y_pred.argmax(dim=self.axis)
        true_class = y_true.long()
        correct_predictions = pred_class == true_class
        loss = correct_predictions.float().mean(dim=self.axis)
        if self.reduction == "none":
            return loss
        elif self.reduction == "sum":
            return loss.sum()
        elif self.reduction == "mean":
            return loss.mean()



"""doc
# leanai.data.dataset

> A generic implementation for a dataset based on parsers and file providers.
"""
from typing import Any, Dict, Iterator, List
from torch.utils.data import IterableDataset as _IterableDataset
from torch.utils.data import Dataset as _Dataset

from .parser import IParser, Parser
from .file_provider import FileProviderSequence, FileProviderIterable
from .data_promise import DataPromise


class IIterableDataset(_IterableDataset):
    """
    Interface for an iterable dataset.

    Has iter and next.
    """
    def __next__(self) -> Any:
        raise NotImplementedError("Must be implemented by subclass.")

    def __iter__(self) -> Iterator[Any]:
        raise NotImplementedError("Must be implemented by subclass.")


class ISequenceDataset(_Dataset):
    """
    Interface for a sequence dataset.

    Has len, getitem, iter and next.
    """
    def __next__(self) -> Any:
        raise NotImplementedError("Must be implemented by subclass.")

    def __iter__(self) -> Iterator[Any]:
        raise NotImplementedError("Must be implemented by subclass.")

    def __getitem__(self, index) -> Any:
        raise NotImplementedError("Must be implemented by subclass.")

    def __len__(self) -> int:
        raise NotImplementedError("Must be implemented by subclass.")


class _CommonDataset:
    def __init__(self, file_provider_iterable: FileProviderIterable, parser: IParser) -> None:
        super().__init__()
        self._file_provider = file_provider_iterable
        self._fp_iterator = None
        self._parser = parser
        self.transformers = []

    def _process(self, sample: Dict[str, DataPromise]) -> Any:
        sample = self._parser(sample)
        for transformer in self.transformers:
            sample = transformer(sample)
        return sample
    
    def __next__(self) -> Any:
        if self._fp_iterator is None:
            raise RuntimeError("You must first call iter(...) before you can use next(...).")
        sample = self._fp_iterator.__next__()
        return self._process(sample)

    def __iter__(self) -> Iterator[Any]:
        self._fp_iterator = self._file_provider.__iter__()
        return self

    def __len__(self) -> int:
        return len(self._file_provider)


class IterableDataset(_CommonDataset, IIterableDataset):
    def __init__(self, file_provider_iterable: FileProviderIterable, parser: IParser) -> None:
        """
        An implementation of the IIterableDataset using fileprovider and parser.

        :param file_provider_iterable: The iterable file provider providing samples to the parser.
        :param parser: The parser converting samples into a usable format.
        """
        super().__init__(file_provider_iterable, parser)


class SequenceDataset(_CommonDataset, ISequenceDataset):
    def __init__(self, file_provider_sequence: FileProviderSequence, parser: IParser) -> None:
        """
        An implementation of the ISequenceDataset using fileprovider and parser.

        :param file_provider_sequence: The sequence file provider providing samples to the parser.
        :param parser: The parser converting samples into a usable format.
        """
        super().__init__(file_provider_sequence, parser)
        self._file_provider = file_provider_sequence

    def __getitem__(self, index) -> Any:
        sample = self._file_provider[index]
        return self._process(sample)

class SimpleDataset(Parser, SequenceDataset):
    def __init__(self, InputType, OutputType) -> None:
        """
        SimpleDataset decodes the samples required to populate the input and output type automatically.

        The SimpleDataset should never be instantiated directly.
        You should inherit it and then instantiate the inherited class.

        The SimpleDataset automatically only parses those examples which are used in the InputType and OutputType.
        Data which is not used in any of the two will not be parsed thus speeding up the parsing process.

        To achieve this, the parser tries to call functions called "parse_{attribute_name}",
        where "attribute_name" is the field name in the named tuple.

        For example providing this named tuple:
        `InputType = NamedTuple("InputType", image=np.ndarray)`
        will result in a call to `parse_image` or an error if that function does not exist.
        The following signature will be expected:
        ```
        def parse_image(self, sample_token) -> np.ndarray:
        ```
        If that is not the case, your code may break somewhere.

        **Arguments**
        :param InputType: A definition of a named tuple that defines the input of the neural network.
        :param OutputType: A definition of a named tuple that defines the output of the neural network.
        """
        Parser.__init__(self, InputType, OutputType)
        SequenceDataset.__init__(self, [], self)

    def set_sample_tokens(self, sample_tokens: List[Any]) -> None:
        """
        Set the list of sample tokens.

        :param sample_tokens: A list of all sample tokens of the dataset.
        """
        self._file_provider = list(sample_tokens)



"""doc
# leanai.data.datasets.pickle_dataset

> An implementation of a cached dataset.
"""
import os
import pickle
from typing import Any, Dict
from leanai.data.parser import IParser
from leanai.data.data_promise import DataPromise, DataPromiseFromBytes
from leanai.data.file_provider import FileProviderSequence
from leanai.data.dataset import SequenceDataset


def _pickle_file_path(cache_path: str, split: str, idx: int):
    assert cache_path is not None
    return os.path.join(cache_path, f"{split}_{idx:09d}.pk")


class PickledDataset(SequenceDataset):
    def __init__(self, split: str, cache_path: str, shuffle: bool):
        """
        Create a dataset from previously pickled examples.

        Files are assumed to be stored as f"{cache_path}/{split}_{idx:09d}.pk".
        
        :param split: The datasplit to load.
        :param cache_path: The path where the data was cached.
        :param shuffle: If the data should be shuffled when iterating over it.
        """
        super().__init__(
            file_provider_sequence=_PickleFileProvider(split, cache_path, shuffle),
            parser=_PickleFileParser(),
        )

    @staticmethod
    def create(data, split: str, cache_path: str):
        """
        Create a pickle dataset on the disk.

        (Use this to create the dataset that later can be loaded via the constructor.)
        
        Files are stored as f"{cache_path}/{split}_{idx:09d}.pk".

        :param data: The dataset to pickle (must be iterable).
        :param split: The datasplit that is stored (used later when loading the data).
        :param cache_path: The path where to store the data cache.
        """
        if not os.path.exists(cache_path):
            os.makedirs(cache_path)
        for idx, sample in enumerate(data):
            fp = _pickle_file_path(cache_path, split, idx)
            with open(fp, "wb") as f:
                pickle.dump(sample, f)


class _PickleFileProvider(FileProviderSequence):
    def __init__(self, split: str, cache_path: str, shuffle: bool) -> None:
        super().__init__(shuffle=shuffle)
        self.cache_path = cache_path
        self.split = split
        self._cache_indices = {}

        if not os.path.exists(cache_path):
            raise FileNotFoundError(f"Cannot find cache at path: {cache_path}")
        # Read all files in the folder into a dict that maps indices to filenames (for quicker access)
        cache_files = os.listdir(cache_path)
        for cf in cache_files:
            if cf.endswith(".pk") and cf.startswith(f"{self.split}_"):
                self._cache_indices[int(cf.replace(".pk", "").replace(f"{self.split}_", ""))] = os.path.join(cache_path, cf)
        self._cached_len = len(self._cache_indices.keys())

    def __len__(self) -> int:
        return self._cached_len

    def __getitem__(self, idx: int) -> Dict[str, DataPromise]:
        with open(_pickle_file_path(self.cache_path, self.split, idx), "rb") as f:
            return {
                "pk": DataPromiseFromBytes(f.read())
            }


class _PickleFileParser(IParser):
    def __call__(self, sample: Dict[str, DataPromise]) -> Any:
        return pickle.loads(sample["pk"].data)



"""doc
# leanai.model.layers.roi_ops

> Operations for region of interest extraction.
"""
import torch
from torch.nn import Module
from torchvision.ops import roi_pool as _roi_pool
from torchvision.ops import roi_align as _roi_align
from leanai.model.module_registry import add_module


def _convert_boxes_to_roi_format(boxes):
    """
    Convert rois into the torchvision format.

    :param boxes: The roi boxes as a native tensor[B, K, 4].
    :return: The roi boxes in the format that roi pooling and roi align in torchvision require. Native tensor[B*K, 5].
    """
    concat_boxes = boxes.view((-1, 4))
    ids = torch.full_like(boxes[:, :, :1], 0)
    for i in range(boxes.shape[0]):
        ids[i, :, :] = i
    ids = ids.view((-1, 1))
    rois = torch.cat([ids, concat_boxes], dim=1)
    return rois


@add_module()
class BoxToRoi(Module):
    def __init__(self, feature_map_scale=1) -> None:
        super().__init__()
        self.feature_map_scale = feature_map_scale
    
    def forward(self, boxes):
        # Compute corners of rois
        min_corner = boxes[:, :2, :] - boxes[:, 2:, :] / 2
        max_corner = boxes[:, :2, :] + boxes[:, 2:, :] / 2
        corners = torch.cat([min_corner, max_corner], dim=1)
        return corners / self.feature_map_scale


@add_module()
class RoiPool(Module):
    def __init__(self, output_size, spatial_scale=1.0):
        """
        Performs Region of Interest (RoI) Pool operator described in Fast R-CNN.

        Creates a callable object, when calling you can use these Arguments:
        * **features**: (Tensor[N, C, H, W]) input tensor
        * **rois**: (Tensor[N, 4, K]) the box coordinates in (cx, cy, w, h) format where the regions will be taken from.
        * **return**: (Tensor[N, C * output_size[0] * output_size[1], K]) The feature maps crops corresponding to the input rois.
        
        Parameters to RoiPool constructor:
        :param output_size: (Tuple[int, int]) the size of the output after the cropping is performed, as (height, width)
        :param spatial_scale: (float) a scaling factor that maps the input coordinates to the box coordinates. Default: 1.0
        """
        super().__init__()
        self.output_size = output_size
        self.spatial_scale = spatial_scale
        
        
    def forward(self, features, rois):
        rois = rois.transpose(1, 2).contiguous()
        torchvision_rois = _convert_boxes_to_roi_format(rois)

        result = _roi_pool(features, torchvision_rois, self.output_size, self.spatial_scale)

        # Fix output shape
        N, C, _, _ = features.shape
        result = result.view((N, -1, C * self.output_size[0]* self.output_size[1]))
        result = result.transpose(1, 2).contiguous()
        return result


@add_module()
class RoiAlign(Module):
    def __init__(self, output_size, spatial_scale=1.0):
        """
        Performs Region of Interest (RoI) Align operator described in Mask R-CNN.

        Creates a callable object, when calling you can use these Arguments:
        * **features**: (Tensor[N, C, H, W]) input tensor
        * **rois**: (Tensor[N, 4, K]) the box coordinates in (x1, y1, x2, y2) format where the regions will be taken from.
        * **return**: (Tensor[N, C * output_size[0] * output_size[1], K]) The feature maps crops corresponding to the input rois.
        
        Parameters to RoiAlign constructor:
        :param output_size: (Tuple[int, int]) the size of the output after the cropping is performed, as (height, width)
        :param spatial_scale: (float) a scaling factor that maps the input coordinates to the box coordinates. Default: 1.0
        """
        super().__init__()
        self.output_size = output_size
        self.spatial_scale = spatial_scale
        
    def forward(self, features, rois):
        rois = rois.transpose(1, 2).contiguous()
        torchvision_rois = _convert_boxes_to_roi_format(rois)

        # :param aligned: (bool) If False, use the legacy implementation.
        #    If True, pixel shift it by -0.5 for align more perfectly about two neighboring pixel indices.
        #    This version in Detectron2
        result = _roi_align(features, torchvision_rois, self.output_size, self.spatial_scale, aligned=True)

        # Fix output shape
        N, C, _, _ = features.shape
        result = result.view((N, -1, C * self.output_size[0]* self.output_size[1]))
        result = result.transpose(1, 2).contiguous()
        return result



"""doc
# leanai.model.layers.selection

> These layers select parts of a tensor.
"""
import torch
from torch.nn import Module
from leanai.model.module_registry import add_module


@add_module()
class Gather(Module):
    def __init__(self, axis):
        """
        Gather tensors from one tensor by providing an index tensor.

        ```
        assert src.shape = [B, X, Y, Z]
        assert idx.shape = [B, K]
        assert 0 <= idx.min() and idx.max() < src.shape[axis]
        # -->
        assert Gather(1)(src, idx).shape  == [B, K, Y, Z]
        assert Gather(2)(src, idx).shape  == [B, X, K, Z]
        assert Gather(3)(src, idx).shape  == [B, X, Y, K]
        #assert Gather(0) -> Exception
        ```
    
        Created object is callable with the following parameters:
        * **input_tensor**: (Tensor[B, ..., L, ...]) The tensor from which to gather values at the given indices.
        * **indices**: (Tensor[B, K]) The indices at which to return the values of the input tensor.
        * **returns**: (Tensor[B, ..., K, ...]) The tensor containing the values at the indices given.

        Arguments:
        :param axis: The axis along which to select.
        """
        super().__init__()
        assert axis != 0, "You cannot gather over the batch dimension."
        self.axis = axis

    def forward(self, input_tensor, indices):
        import torch
        assert input_tensor.shape[0] == indices.shape[0]
        assert len(indices.shape) == 2, "Indices must be of shape (B, K). Found shape: {}".format(indices.shape)
        assert 0 <= indices.min(), "Indices contain values out of bounds. Min idx: {}".format(indices.min())
        assert indices.max() < input_tensor.shape[self.axis], "Indices contain values out of bounds. Max idx: {}, Shape: {}, Axis: {}".format(
            indices.max(), input_tensor.shape, self.axis)

        # Then gather the indices along the batches.
        batchless_axis = self.axis - 1 if self.axis > 0 else self.axis
        return torch.stack([torch.index_select(input_tensor[i], batchless_axis, indices[i]) for i in range(indices.shape[0])])


@add_module()
class TopKIndices(Module):
    def __init__(self, k):
        """
        Returns the top k tensor indices (separate per batch).
    
        Created object is callable with the following parameters:
        * **input_tensor**: (Tensor[N, L]) The tensor in which to search the top k indices.
        * **returns**: (Tensor[N, K]) The tensor containing the indices of the top k values.
        
        Parameters for the constructor:
        :param k: The number of indices to return per batch.
        """
        super().__init__()
        self.k = k

    def forward(self, input_tensor):
        return torch.topk(input_tensor, self.k).indices


@add_module()
class GatherTopKIndices(Module):
    def __init__(self, k: int, background_class_idx: int = 0):
        """
        Returns the top k tensor indices (separate per batch).
    
        For shapes: B=#Batches, X=Arbitrary, C=#Classes, N=#Samples.

        Created object is callable with the following parameters:
        * **input_tensor**: (Tensor[B, X, N]) The tensor from which to gather the values.
        * **scores**: (Tensor[B, C, N]) The tensor in which to search the top k indices.
        * **returns**: (Tensor[B, X, k]) The tensor containing the values at the top k indices.
        
        Parameters for the constructor:
        :param k: The number of indices to return per batch.
        :param background_class_idx: (int) The index at which the background class is. (Default: 0)
        """
        super().__init__()
        self.gather = Gather(axis=-1)
        self.topk = TopKIndices(k)
        self.background_class_idx = background_class_idx

    def forward(self, input_tensor, scores):
        foreground_scores = 1 - scores[:, self.background_class_idx]
        indices = self.topk(foreground_scores)
        return self.gather(input_tensor, indices)



"""doc
# leanai.data.datasets.fashion_mnist_dataset

> An implementation of a dataset for fashion mnist.
"""
from typing import Any, NamedTuple, Tuple
import numpy as np
from torchvision.datasets import FashionMNIST
from leanai.core.definitions import SPLIT_TRAIN
from leanai.data.dataset import SimpleDataset


MNISTInputType = NamedTuple("MNISTInput", image=np.ndarray)
MNISTOutputType = NamedTuple("MNISTOutput", class_id=np.ndarray)


class FashionMNISTDataset(SimpleDataset):
    def __init__(self, split: str, data_path: str = "", download=True) -> None:
        super().__init__(MNISTInputType, MNISTOutputType)
        self.dataset = FashionMNIST(data_path, train=split == SPLIT_TRAIN, download=download)
        self.set_sample_tokens(range(len(self.dataset)))

    def get_image(self, sample) -> np.ndarray:
        image = np.array(self.dataset[sample][0], dtype="float32")
        return np.reshape(image, (28, 28, 1))

    def get_class_id(self, sample) -> np.ndarray:
        return np.array([self.dataset[sample][1]], dtype="uint8")


def test_visualization(data_path):
    import matplotlib.pyplot as plt
    dataset = FashionMNISTDataset(SPLIT_TRAIN, data_path)
    networkInput, networkOutput = dataset[0]
    plt.title(networkOutput.class_id)
    plt.imshow(networkInput.image)
    plt.show()


if __name__ == "__main__":
    import sys
    data_path = sys.argv[1] if len(sys.argv) > 1 else "test_logs/FashionMNIST"
    test_visualization(data_path)



"""doc
# leanai.data.transformer

> A transformer converts the data from a general format into what the neural network needs.
"""


class Transformer():
    def __init__(self):
        """
        A transformer must implement `__call__`.
        
        A transformer is a callable that gets the data (usually a tuple of feature, label), transforms it and returns the data again (usually as a tuple of feature, label).
        The last transformer must output a tuple of feature of type NetworkInput (namedtuple) and label of type NetworkOutput (namedtuple) for babilim to be able to pass it to the neural network.
        """
        pass

    def __call__(self, *args):
        """
        This function gets the data from the previous transformer or dataset as input and should output the data again.

        :param args: The input data.
        :return: The output data.
        """
        raise NotImplementedError

    @property
    def version(self):
        """
        Defines the version of the transformer. The name can be also something descriptive of the method.

        :return: The version number of the transformer.
        """
        raise NotImplementedError



"""doc
# leanai.model.layers.batch_normalization

> Apply batch normalization to a tensor.
"""
import torch
from torch.nn import BatchNorm1d, BatchNorm2d, BatchNorm3d
from torch.nn import Module
from leanai.core.annotations import RunOnlyOnce
from leanai.model.module_registry import add_module


@add_module()
class BatchNormalization(Module):
    def __init__(self):
        """
        A batch normalization layer.
        """
        super().__init__()
        
    @RunOnlyOnce
    def build(self, features):
        if len(features.shape) == 2 or len(features.shape) == 3:
            self.bn = BatchNorm1d(num_features=features.shape[1])
        elif len(features.shape) == 4:
            self.bn = BatchNorm2d(num_features=features.shape[1])
        elif len(features.shape) == 5:
            self.bn = BatchNorm3d(num_features=features.shape[1])
        else:
            raise RuntimeError("Batch norm not available for other input shapes than [B,L], [B,C,L], [B,C,H,W] or [B,C,D,H,W] dimensional.")
        
        if torch.cuda.is_available():
            self.bn = self.bn.to(torch.device(features.device))

    def forward(self, features):
        self.build(features)
        return self.bn(features)



"""doc
# leanai.core.cli_tools

> The command line interface for project management with leanai.

You can create a new project with the following command:
```bash
leanai --new_project --name=<name>
```

Also you can add run configurations for vscode for a project with this command:
```bash
leanai --vscode_run_config <name> <module> <arg1> <arg2> ...
```
"""
import sys
import os
import json
import jstyleson
import argparse
import shutil
TEMPLATE_CLONE_URL = "https://github.com/penguinmenac3/leanai-template.git"


def main():
    args = sys.argv[1:]

    valid_tools = ["--new_project", "--vscode_run_config"]
    if args[0] not in valid_tools:
        print("Unknown tool {}. Chose one in {}.".format(args[0], valid_tools))
        return

    if args[0] == "--new_project":
        new_project(args[1:])
    elif args[0] == "--vscode_run_config":
        vscode_run_config(args[1:])


def new_project(args):
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('name', type=str, help="Project name")
    args, run_args = parser.parse_known_args(args)
    if os.path.exists(args.name):
        print("Cannot create project. Folder '{}' already exists.".format(args.name))
        return
    os.system("git clone {} {}".format(TEMPLATE_CLONE_URL, args.name))
    os.chdir(args.name)
    shutil.move("template", args.name)
    vscode_run_config(["Train", "{}.config.train".format(args.name)])
    vscode_run_config(["Test", "{}.config.test".format(args.name), "--mode=test"])


def vscode_run_config(args):
    parser = argparse.ArgumentParser(description='Process some integers.')
    parser.add_argument('name', type=str, help="The name under which it should show in the run config dropdown.")
    parser.add_argument('module', type=str, help="The name of the module that should be executed, e.g. 'leanai.core.cli'")
    args, run_args = parser.parse_known_args(args)
    if not os.path.exists(".vscode"):
        os.mkdir(".vscode")
        print("Created .vscode folder.")
    if not os.path.exists(os.path.join(".vscode", "launch.json")):
        with open(os.path.join(".vscode", "launch.json"), "w") as f:
            data = {"version": "0.2.0","configurations": []}
            f.write(json.dumps(data, indent=4, sort_keys=True))
            f.write("\n")
        print("Created launch.json.")
    with open(os.path.join(".vscode", "launch.json"), "r") as f:
        data = jstyleson.load(f)
        data["configurations"].append({
            "name": args.name,
            "type": "python",
            "request": "launch",
            "module": args.module,
            "args": run_args
        })
    with open(os.path.join(".vscode", "launch.json"), "w") as f:
        f.write(json.dumps(data, indent=4, sort_keys=True))
        f.write("\n")
    print("Added configuration under name: {}. You may select it in the debug menu now.".format(args.name))




"""doc
# leanai.training.losses.regression

> All losses related to regression problems.
"""
from torch.nn import Module
from torch.nn import SmoothL1Loss as _SmoothL1Loss
from torch.nn import MSELoss as _MSELoss


class SmoothL1Loss(Module):
    def __init__(self, model=None, reduction: str = "mean"):
        """
        Compute a smooth l1 loss.
        
        :param reduction: Specifies the reduction to apply to the output: `'none'` | `'mean'` | `'sum'`. `'none'`: no reduction will be applied, `'mean'`: the sum of the output will be divided by the number of elements in the output, `'sum'`: the output will be summed. Default: 'mean'.
        """
        super().__init__()
        self.loss_fun = _SmoothL1Loss(reduction=reduction)
        
    def forward(self, y_pred, y_true):
        """
        Compute the smooth l1 loss.
        
        :param y_pred: The predictions of the network as a tensor.
        :param y_true: The desired outputs of the network (labels) as a tensor.
        """
        return self.loss_fun(y_pred, y_true)


class MSELoss(Module):
    def __init__(self, model=None, reduction: str = "mean"):
        """
        Compute a mse loss.
        
        :param reduction: Specifies the reduction to apply to the output: `'none'` | `'mean'` | `'sum'`. `'none'`: no reduction will be applied, `'mean'`: the sum of the output will be divided by the number of elements in the output, `'sum'`: the output will be summed. Default: 'mean'.
        """
        super().__init__()
        self.loss_fun = _MSELoss(reduction=reduction)
        
    def forward(self, y_pred, y_true):
        """
        Compute the mse loss.
        
        :param y_pred: The predictions of the network as a tensor.
        :param y_true: The desired outputs of the network (labels) as a tensor.
        """
        return self.loss_fun(y_pred, y_true)



"""doc
# leanai.training.losses.masking

> Masking losses can be made easy by putting nans or negative values in the ground truth.
"""
import torch
from torch.nn import Module


class MaskedLoss(Module):
    def __init__(self, loss, keep_dim=1):
        """
        """
        super().__init__()
        self.wrapped_loss = loss
        self.keep_dim = keep_dim

    def forward(self, y_pred, y_true):
        """
        Compute the loss given in the constructor only on values where the GT masking fun returns true.
        
        :param y_pred: The predictions of the network. Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        :param y_true: The desired outputs of the network (labels). Either a NamedTuple pointing at ITensors or a Dict or Tuple of ITensors.
        """
        assert len(y_pred.shape) == len(y_true.shape)
        if self.keep_dim is not None:
            keep_dim = self.keep_dim if self.keep_dim >= 0 else len(y_pred.shape) + self.keep_dim
            if keep_dim != len(y_pred.shape) - 1:
                y_true = torch.transpose(y_true, keep_dim, len(y_true.shape) - 1)
                y_pred = torch.transpose(y_pred, keep_dim, len(y_pred.shape) - 1)
        y_pred_N = y_pred.shape[-1]
        y_true_N = y_true.shape[-1]

        binary_mask = self.masking_fun(y_true)
        mask = binary_mask.float()
        masked_y_true = (y_true * mask)[binary_mask]

        for dim in range(len(binary_mask.shape)):
            if y_pred.shape[dim] != binary_mask.shape[dim] and y_pred.shape[dim] % binary_mask.shape[dim] == 0:
                repeat = y_pred.shape[dim] / binary_mask.shape[dim]
                binary_mask = torch.repeat_interleave(binary_mask, int(repeat), dim=dim)
        mask = binary_mask.float()
        masked_y_pred = (y_pred * mask)[binary_mask]
        
        if self.keep_dim is not None:
            masked_y_pred = masked_y_pred.reshape((-1, y_pred_N))
            masked_y_true = masked_y_true.reshape((-1, y_true_N))

        if masked_y_pred.shape[0] > 0:
            loss = self.wrapped_loss(masked_y_pred, masked_y_true)
        else:
            loss = 0

        return loss

    def masking_fun(self, tensor):
        raise RuntimeError("Must be implemented by subclass or set during runtime.")


class NegMaskedLoss(MaskedLoss):
    def __init__(self, loss, keep_dim=1):
        """
        """
        super().__init__(loss=loss, keep_dim=keep_dim)
    
    def masking_fun(self, tensor):
        return tensor >= 0


class ValueMaskedLoss(MaskedLoss):
    def __init__(self, loss, ignore_value, keep_dim=1):
        """
        """
        super().__init__(loss=loss, keep_dim=keep_dim)
        self.ignore_value = ignore_value

    def masking_fun(self, tensor):
        return tensor != self.ignore_value

class NaNMaskedLoss(MaskedLoss):
    def __init__(self, loss, keep_dim=1):
        """
        """
        super().__init__(loss=loss, keep_dim=keep_dim)

    def masking_fun(self, tensor):
        return ~tensor.isnan()



"""doc
# leanai.model.layers.flatten

> Flatten a feature map into a linearized tensor.
"""
from torch.nn import Module

from leanai.model.module_registry import add_module


@add_module()
class Flatten(Module):
    def __init__(self, dims=2):
        """
        Flatten a feature map into a linearized tensor.
    
        This is usefull after the convolution layers before the dense layers. The (B, W, H, C) tensor gets converted ot a (B, N) tensor.
        :param dims: The number of dimensions that should be kept after flattening. Default is 2.
        """
        super().__init__()
        self.dims = dims

    def forward(self, features):
        shape = []
        for i in range(self.dims-1):
            shape.append(features.shape[i])
        shape.append(-1)
        return features.view(*shape)



"""doc
# leanai.model.layers.convolution

> Convolution for 1d, 2d and 3d.
"""
import torch
from torch.nn import Module
from torch.nn import Conv1d as _Conv1d
from typing import Optional, Any, Tuple
from leanai.core.annotations import RunOnlyOnce
from leanai.model.module_registry import add_module


@add_module()
class Conv1D(Module):
    def __init__(self, filters: int, kernel_size: int, padding: Optional[str] = None, stride: int = 1, dilation_rate: int = 1, kernel_initializer: Optional[Any] = None, activation=None):
        """
        A 1d convolution layer.
    
        :param filters: The number of filters in the convolution. Defines the number of output channels.
        :param kernel_size: The kernel size of the convolution. Defines the area over which is convolved. Typically 1, 3 or 5 are recommended.
        :param padding: What type of padding should be applied. The string "none" means no padding is applied, None or "same" means the input is padded in a way that the output stays the same size if no stride is applied.
        :param stride: The offset between two convolutions that are applied. Typically 1. Stride affects also the resolution of the output feature map. A stride 2 halves the resolution, since convolutions are only applied every odd pixel.
        :param dilation_rate: The dilation rate for a convolution.
        :param kernel_initializer: A kernel initializer function. By default orthonormal weight initialization is used.
        :param activation: The activation function that should be added after the dense layer.
        """
        super().__init__()
        self.filters = filters
        self.kernel_size = kernel_size
        self.padding = padding
        self.dilation = dilation_rate
        self.stride = stride
        self.kernel_initializer = kernel_initializer
        self.activation = activation
        
    @RunOnlyOnce
    def build(self, features):
        if self.kernel_initializer is None:
            from torch.nn.init import orthogonal_
            self.kernel_initializer = orthogonal_
        if self.padding == "same" or self.padding is None:
            self.padding = int((self.kernel_size - 1) / 2)
        elif self.padding == "none":
            self.padding = 0
        else:
            raise NotImplementedError("Padding {} is not implemented.".format(self.padding))
        in_channels = features.shape[1]
        self.conv = _Conv1d(in_channels, self.filters, self.kernel_size, self.stride, self.padding, self.dilation)
        self.conv.weight.data = self.kernel_initializer(self.conv.weight.data)
        if torch.cuda.is_available():
            self.conv = self.conv.to(torch.device(features.device))

    def forward(self, features):
        self.build(features)
        result = self.conv(features)
        if self.activation is not None:
            result = self.activation(result)
        return result


@add_module()
class Conv2D(Module):
    def __init__(self, filters: int, kernel_size: Tuple[int, int], padding: Optional[str] = None, strides: Tuple[int, int] = (1, 1), dilation_rate: Tuple[int, int] = (1, 1), kernel_initializer: Optional[Any] = None, activation=None):
        """
        A 2d convolution layer.
    
        :param filters: The number of filters in the convolution. Defines the number of output channels.
        :param kernel_size: The kernel size of the convolution. Defines the area over which is convolved. Typically (1,1) (3,3) or (5,5) are recommended.
        :param padding: What type of padding should be applied. The string "none" means no padding is applied, None or "same" means the input is padded in a way that the output stays the same size if no stride is applied.
        :param stride: The offset between two convolutions that are applied. Typically (1, 1). Stride affects also the resolution of the output feature map. A stride 2 halves the resolution, since convolutions are only applied every odd pixel.
        :param dilation_rate: The dilation rate for a convolution.
        :param kernel_initializer: A kernel initializer function. By default orthonormal weight initialization is used.
        :param activation: The activation function that should be added after the dense layer.
        """
        super().__init__()
        self.filters = filters
        self.kernel_size = kernel_size
        self.padding = padding
        self.dilation = dilation_rate
        self.stride = strides
        self.kernel_initializer = kernel_initializer
        self.activation = activation
        
    @RunOnlyOnce
    def build(self, features):
        import torch
        from torch.nn import Conv2d as _Conv2d
        if self.kernel_initializer is None:
            from torch.nn.init import orthogonal_
            self.kernel_initializer = orthogonal_
        if self.padding == "same" or self.padding is None:
            px = int((self.kernel_size[0] - 1) / 2)
            py = int((self.kernel_size[1] - 1) / 2)
            self.padding = (px, py)
        elif self.padding == "none":
            self.padding = (0, 0)
        else:
            raise NotImplementedError("Padding {} is not implemented.".format(self.padding))
        in_channels = features.shape[1]
        self.conv = _Conv2d(in_channels, self.filters, self.kernel_size, self.stride, self.padding, self.dilation)
        self.conv.weight.data = self.kernel_initializer(self.conv.weight.data)
        if torch.cuda.is_available():
            self.conv = self.conv.to(torch.device(features.device))
        
    def forward(self, features):
        self.build(features)
        result = self.conv(features)
        if self.activation is not None:
            result = self.activation(result)
        return result



from collections import namedtuple
from leanai.core.logging import warn
from typing import Sequence
import torch.nn as nn
from torch import Tensor
from leanai.core import logging
from leanai.model import module_registry
from leanai.model.module_registry import add_lib_from_json
import leanai.model.layers  # required so layers are added to registry
import leanai.model.models  # required so models are added to registry


WARN_DISABLED_LAYERS = True


def _cleaned_spec(spec):
    spec_light = spec.copy()
    if "in" in spec_light:
        del spec_light["in"]
    if "out" in spec_light:
        del spec_light["out"]
    if "type" in spec_light:
        del spec_light["type"]
    if "typedef" in spec_light:
        del spec_light["typedef"]
    return spec_light


def get_val_from_spec(val, spec):
    tokens = val.replace("spec:", "").split("=")
    spec_name = tokens[0]
    if spec_name in spec:
        return spec[spec_name]
    else:
        if len(tokens) == 1:
            return False
        else:
            if tokens[1] == "True" or tokens[1] == "true":
                return True
            elif tokens[1] == "False" or tokens[1] == "false":
                return False
            elif tokens[1].startswith("\"") or tokens[1].startswith("'"):
                return tokens[1][1:-1]
            else:
                try:
                    return float(tokens[1])
                except:
                    raise RuntimeError(f"Invalid value must be true, false, a number or a string (enclosed in quotation marks). Found: {tokens[1]}")


class Module(nn.Module):
    @staticmethod
    def create(typename, _local_variables={}, **spec):
        if "disabled" in spec:
            del spec["disabled"]

        # In case the module is in the json module library update it
        while typename in module_registry._json_module_library:
            _local_variables = {}  # Cut variable passing for typedefs
            spec.update(**module_registry._json_module_library[typename])
            if "typedef" not in spec:
                spec["typedef"] = typename
            typename = spec["type"]
        if "typedef" not in spec:
            spec["typedef"] = typename

        # If it is a composite module, create it.
        if typename in ["Sequential", "Paralell"]:
            args = spec["args"] if "args" in spec else ["input"]
            returns = spec["return"] if "return" in spec else [None]
            if typename == "Sequential":
                module = Sequential(args, returns, spec["layers"], spec, _local_variables)
                class _new_seq_class(Sequential):
                    pass
                _new_seq_class.__name__ = spec["typedef"]
                module.__class__ = _new_seq_class
            else:
                module = Paralell(args, returns, spec["layers"], spec, _local_variables)
                class new_par_class(Paralell):
                    pass
                new_par_class.__name__ = spec["typedef"]
                module.__class__ = new_par_class
        else:
            if typename in module_registry._native_module_library:
                spec_light = spec.copy()
                spec_light = _cleaned_spec(spec)
                module = module_registry._native_module_library[typename](**spec_light)
            else:
                raise RuntimeError("There is no layer for '{}' in the library!".format(typename))
        module._module_inputs = spec["in"] if "in" in spec else [None]
        if isinstance(module._module_inputs, str):
            module._module_inputs = [module._module_inputs]
        module._module_outputs = spec["out"] if "out" in spec else []  # First output is always stored to -1 so no need to specify here
        if isinstance(module._module_outputs, str):
            module._module_outputs = [module._module_outputs]
        module._spec = spec
        return module

    @staticmethod
    def create_from_file(filename, typename, _local_variables={}, **spec):
        add_lib_from_json(filename)
        return Module.create(typename, _local_variables=_local_variables, **spec)

    def __init__(self, args, returns, layers, spec, _local_variables):
        super().__init__()
        self._args = args
        self._returns = returns
        can_namedtuple_output = True
        for ret_name in returns:
            if not isinstance(ret_name, str):
                can_namedtuple_output = False
                break
        self._return_type = None
        if can_namedtuple_output and len(returns) > 1:
            self._return_type = namedtuple("ReturnType", self._returns)
        self.submodules = []
        self._inputs = []
        self._outputs = []
        self._spec = spec
        for idx, layer in enumerate(layers):
            layer = layer.copy()
            for key, val in layer.items():
                # Allow or for bool operations
                if isinstance(val, str) and val.startswith("spec:") and "|" in val:
                    result = False
                    for v in val.split("|"):
                        result = result or get_val_from_spec(v, self._spec)
                    layer[key] = result
                # Use value directly if no or in value spec.
                else:
                    if isinstance(val, str) and val.startswith("spec:"):
                        layer[key] = get_val_from_spec(val, self._spec)
            if "disabled" in layer and layer["disabled"]:
                if WARN_DISABLED_LAYERS:
                    warn("Disabled Layer: {}".format(layer))
                continue
            layer_type = layer["type"]
            del layer["type"]
            name = layer["name"] if "name" in layer else "layer_{}".format(idx)
            if "name" in layer:
                del layer["name"]
            module = Module.create(layer_type, _local_variables=_local_variables, **layer)
            self.submodules.append(module)
            self.add_module(name, module)
        self._local_variables = _local_variables

    def forward(self, *args):
        self._store_variables(self._args, args)
        self.forward_impl()
        returns = self._collect_variables(self._returns, no_tuple_for_single_output=True)
        if self._return_type is not None:
            if "allow_return_none" not in self._spec or not self._spec["allow_return_none"]:
                for idx, ret in enumerate(returns):
                    if ret is None:
                        raise RuntimeError(f"None value at index {idx} in output of module. If this was intentional add the 'allow_return_none' to the spec.")
            returns = self._return_type(*returns)
        return returns

    def forward_impl(self):
        raise NotImplementedError("Must be implemented by subclasses.")

    def _collect_variables(self, names, no_tuple_for_single_output=False):
        collected = []
        for key in names:
            collected.append(self._local_variables[key])
        # Unpack single returns to be no tuple
        if no_tuple_for_single_output and len(collected) == 1:
            collected = collected[0]
        return collected

    def _store_variables(self, names, values):
        assert isinstance(values, Sequence) and not isinstance(values, Tensor)
        for key, value in zip(names, values):
            self._local_variables[key] = value
        self._local_variables[None] = values[0]


class Sequential(Module):
    def __init__(self, args, returns, layers, spec, _local_variables):
        super().__init__(args, returns, layers, spec, _local_variables)

    def forward_impl(self):
        for layer in self.submodules:
            if layer is not None:
                layer_args = self._collect_variables(layer._module_inputs)
                if logging.DEBUG_VERBOSITY:
                    logging.debug(type(layer).__name__)
                    for inp in layer_args:
                        logging.debug(f"INPUT {list(inp.shape)}, {inp.min():.3f} <= inp <= {inp.max():.3f}")
                try:
                    result = layer(*layer_args)
                    if "allow_return_none" not in layer._spec or not layer._spec["allow_return_none"]:
                        for idx, ret in enumerate(result):
                            if ret is None:
                                raise RuntimeError(f"None value at index {idx} in output of module. If this was intentional add the 'allow_return_none' to the spec.")
                except Exception as e:
                    raise type(e)(f"{type(layer).__name__}: {e}") from e
                if not isinstance(result, Sequence) or isinstance(result, Tensor):
                    result = [result]
                if logging.DEBUG_VERBOSITY:
                    logging.debug(type(layer).__name__)
                    for inp in result:
                        logging.debug(f"OUTPUT {list(inp.shape)}, {inp.min():.3f} <= outp <= {inp.max():.3f}")
                self._store_variables(layer._module_outputs, result)


class Paralell(Module):
    def __init__(self, args, returns, layers, spec, _local_variables):
        super().__init__(args, returns, layers, spec, _local_variables)

    def forward_impl(self):
        raise NotImplementedError("Module for paralell not implemented yet.")



"""doc
# leanai.data.datasets.images_in_classfolders_dataset

> An implementation of a dataset where the images are stored in folders which have the class names.
"""
from typing import NamedTuple
import numpy as np
import os
import cv2
from leanai.core.definitions import SPLIT_TRAIN, SPLIT_VAL, SPLIT_TEST
from leanai.data.dataset import SimpleDataset


InputType = NamedTuple("Input", image=np.ndarray)
OutputType = NamedTuple("Output", class_id=np.ndarray)


class ImagesInClassfoldersDataset(SimpleDataset):
    def __init__(self, split: str, data_path, data_train_split=0.6, data_val_split=0.2, data_test_split=0.2) -> None:
        super().__init__(InputType, OutputType)
        self.data_path = data_path
        self.classes = os.listdir(data_path)
        sample_tokens = []

        self.class_mapping = {}
        for idx, class_id in enumerate(self.classes):
            self.class_mapping[class_id] = idx

        # Split the data
        assert data_train_split + data_val_split + data_test_split == 1
        train_end = int(data_train_split*len(self.classes))
        val_end = int((data_train_split+data_val_split)*len(self.classes))
        if split == SPLIT_TRAIN:
            self.classes = self.classes[:train_end]
        if split == SPLIT_VAL:
            self.classes = self.classes[train_end:val_end]
        if split == SPLIT_TEST:
            self.classes = self.classes[val_end:]

        for class_id in self.classes:
            image_names = os.listdir(os.path.join(data_path, class_id))
            image_names = [class_id + "/" + image_name for image_name in image_names]
            sample_tokens.extend(image_names)
        self.set_sample_tokens(sample_tokens)

    def get_image(self, sample_token):
        image = cv2.imread(os.path.join(self.data_path, sample_token))[:,:,::-1]
        return image

    def get_class_id(self, sample_token):
        label = sample_token.split("/")[0]
        return label


def test_visualization(data_path):
    import matplotlib.pyplot as plt
    dataset = ImagesInClassfoldersDataset(SPLIT_TRAIN, "")
    image, class_id = dataset[0]
    plt.title(class_id.class_id)
    plt.imshow(image[0])
    plt.show()


if __name__ == "__main__":
    import sys
    test_visualization(sys.argv[1])



"""doc
# leanai.model.models.imagenet

> An implemntation of various imagenet models.
"""
import torch
from torch.nn import Module
from torchvision.models import vgg16, vgg16_bn, vgg19, vgg19_bn, resnet50, resnet101, resnet152, densenet121, densenet169, densenet201, inception_v3, mobilenet_v2
from torch.nn import Sequential

from leanai.core.annotations import RunOnlyOnce
from leanai.model.layers import ImageConversion
from leanai.model.module_registry import add_module


@add_module("ImageNetModel")
class ImagenetModel(Module):
    def __init__(self, encoder_type, only_encoder=False, pretrained=True, last_layer=None, standardize=True, to_channel_first=True):
        """
        Create one of the iconic image net models in one line.
        Allows for only using the encoder part.

        This model assumes the input image to be 0-255 (8 bit integer) with 3 channels.

        :param encoder_type: The encoder type that should be used. Must be in ("vgg16", "vgg16_bn", "vgg19", "vgg19_bn", "resnet50", "resnet101", "resnet152", "densenet121", "densenet169", "densenet201", "inception_v3", "mobilenet_v2")
        :param only_encoder: Leaves out the classification head for VGG16 leaving you with a feature encoder.
        :param pretrained: If you want imagenet weights for this network.
        :param last_layer: Index of the last layer in the encoder. Allows to cutoff encoder after a few layers.
        :param standardize: (bool) If this is enabled the module will normalize a 0-255 uint image into a float32 image which is normalized and standardized, as with the magic values in pytorch. (Makes them compatible with the imagenet models.)
        :param to_channel_first: (bool) It this is enabled, the image will be converted from h,w,c into the c,h,w format which pytorch uses.
        """
        super().__init__()
        self.only_encoder = only_encoder
        self.pretrained = pretrained
        self.encoder_type = encoder_type
        self.last_layer = last_layer
        self.image_conversion = ImageConversion(standardize=standardize, to_channel_first=to_channel_first)

    @RunOnlyOnce
    def build(self, image):
        model = None
        if self.encoder_type == "vgg16":
            model = vgg16
        elif self.encoder_type == "vgg16_bn":
            model = vgg16_bn
        elif self.encoder_type == "vgg19":
            model = vgg19
        elif self.encoder_type == "vgg19_bn":
            model = vgg19_bn
        elif self.encoder_type == "resnet50":
            model = resnet50
        elif self.encoder_type == "resnet101":
            model = resnet101
        elif self.encoder_type == "resnet152":
            model = resnet152
        elif self.encoder_type == "densenet121":
            model = densenet121
        elif self.encoder_type == "densenet169":
            model = densenet169
        elif self.encoder_type == "densenet201":
            model = densenet201
        elif self.encoder_type == "inception_v3":
            model = inception_v3
        elif self.encoder_type == "mobilenet_v2":
            model = mobilenet_v2
        else:
            raise RuntimeError("Unsupported encoder type.")
        
        if self.only_encoder:
            encoder = list(model(pretrained=self.pretrained).features)
            if self.last_layer is not None:
                encoder = encoder[:self.last_layer+1]
            self.model = Sequential(*encoder)
        else:
            self.model = model(pretrained=self.pretrained)
        
        if torch.cuda.is_available():
            self.model = self.model.to(torch.device(image.device))

    def forward(self, image):
        image = self.image_conversion(image)
        self.build(image)
        return self.model(image)



"""doc
# leanai.model.layers.box_tools

> Convert the output of a layer into a box by using the anchor box.
"""
import torch
import math
import numpy as np
from torch.nn import Module
from leanai.core.annotations import RunOnlyOnce
from leanai.model.module_registry import add_module
from leanai.model.layers.convolution import Conv1D


@add_module()
class DetectionHead(Module):
    def __init__(self, num_classes):
        """
        A detection head module.

        :param num_classes: The number of classes that should be predicted (without softmax). If <= 0 is provided then no class is predicted and None returned as class id.
        """
        super().__init__()
        self.num_classes = num_classes

    @RunOnlyOnce
    def build(self, features, anchors):
        # assume shape of anchors = (batch, 4/6, num_anchors, N)
        dim = self._get_dim(anchors)
        num_anchors = self._get_num_anchors(anchors)
        self.deltas = Conv1D(filters=num_anchors*dim*2, kernel_size=1)
        if self.num_classes > 0:
            self.classification = Conv1D(filters=num_anchors*self.num_classes, kernel_size=1)

    def _get_dim(self, anchors):
        return anchors.shape[1] // 2  # as there is position and size per dimension

    def _get_num_anchors(self, anchors):
        return anchors.shape[2]

    def _normalize_inputs(self, features, anchors):
        features_in_shape = features.shape
        anchors_in_shape = anchors.shape
        if len(features_in_shape) == 4: # BCHW
            assert len(anchors.shape) == 5, "The shape of the anchors must be BCAHW matching the BCHW pattern of the features."
            features = features.reshape(features_in_shape[0], features_in_shape[1], -1)
            anchors = anchors.reshape(anchors_in_shape[0], anchors_in_shape[1], anchors_in_shape[2], -1)
        elif len(features_in_shape) == 3: # BCN
            assert len(anchors.shape) == 3, "The shape of the anchors must match the BCN pattern of the features."
            anchors = anchors.reshape(anchors_in_shape[0], anchors_in_shape[1], 1, -1)
        else:
            raise RuntimeError("Box tensor has an incompatible shape: {}".format(features_in_shape))
        assert len(features.shape) == 3
        return features, anchors

    def _forward_class_ids(self, features):
        class_ids = None
        if self.num_classes > 0:
            class_ids = self.classification(features)
            class_ids = class_ids.reshape((class_ids.shape[0], self.num_classes, -1))
        return class_ids

    def _forward_boxes(self, features, anchors):
        deltas = self.deltas(features)
        deltas = deltas.reshape((deltas.shape[0], self._get_dim(anchors)*2, -1))
        return deltas

    def forward(self, features, anchors):
        """
        Compute the detections given the features and anchors.

        For shape definitions: B=Batchsize, C=#Channels, A=#Anchors, H=Height, W=Width, N=#Boxes.
        Please note, that the shapes must be either of group (a) or of group (b) for all parameters.

        :param features: The feature tensor of shape (a) "BCHW" or (b) "BCN".
        :param anchors: The anchor tensor of shape (a) "BCAHW" or (b) "BCN".
        :returns: The predicted boxes of shape (a) "BC(AWH)" or (b) "BCN". Note how N = (AWH) in the output, resulting in len(shape) == 3 in both cases.
        """
        features, anchors = self._normalize_inputs(features, anchors)
        self.build(features, anchors)
        return self._forward_boxes(features, anchors), self._forward_class_ids(features)


@add_module()
class LogDeltasToBoxes(Module):
    def __init__(self) -> None:
        super().__init__()

    def forward(self, deltas, anchors):
        dim = anchors.shape[1] // 2
        anchor_pos = anchors[:, :dim]
        anchor_size = anchors[:, dim:]
        delta_pos = deltas[:, :dim]
        delta_size = deltas[:, dim:]

        pos = (delta_pos * anchor_size) + anchor_pos
        size = torch.exp(delta_size) * anchor_size
        boxes = torch.cat([pos, size], dim=1)
        return boxes


@add_module()
class DeltasToBoxes(Module):
    def __init__(self, log_deltas=True) -> None:
        super().__init__()
        self.log_deltas = log_deltas

    def forward(self, deltas, anchors):
        dim = anchors.shape[1] // 2
        anchor_pos = anchors[:, :dim]
        anchor_size = anchors[:, dim:]
        delta_pos = deltas[:, :dim]
        delta_size = deltas[:, dim:]

        if not self.log_deltas:
            pos = delta_pos + anchor_pos
            size = delta_size + anchor_size
        else:
            pos = (delta_pos * anchor_size) + anchor_pos
            size = torch.exp(delta_size) * anchor_size
        boxes = torch.cat([pos, size], dim=1)
        return boxes


@add_module()
class GridAnchorGenerator(Module):
    def __init__(self, ratios, scales, feature_map_scale, height=-1, width=-1, base_size=256):
        """
        Construct an anchor grid.

        width = scale * sqrt(ratio) * base_size
        height = scale / sqrt(ratio) * base_size

        :param ratios: A list of aspect ratios used for the anchors.
        :param scales: A list of scales used for the anchors.
        :param feature_map_scale: Divide any meassure in the input space by this number to get the size in the feature map (typically 8, 16 or 32).
        :param height: The height of the anchor grid (in grid cells). When negative will use feature map to figure out size. (Default: -1)
        :param width: The width of the anchor grid (in grid cells). When negative will use feature map to figure out size. (Default: -1)
        :param base_size: The base size of the boxes (defaults to 256).
        """
        super().__init__()
        self.ratios = ratios
        self.scales = scales
        self.feature_map_scale = feature_map_scale
        self.height, self.width = height, width
        self.base_size = base_size

    def _build_anchor_shapes(self, features):
        anchor_shapes = []
        for scale in self.scales:
            for ratio in self.ratios:
                ratio_sqrts = math.sqrt(ratio)
                height = scale / ratio_sqrts * self.base_size
                width = scale * ratio_sqrts * self.base_size
                size = [width, height]
                anchor_shapes.append(size)
        return anchor_shapes

    def _get_anchor_grid_size(self, features):
        _,_,h_feat, w_feat = features.shape
        height = self.height if self.height >= 0 else h_feat
        width = self.width if self.width >= 0 else w_feat
        return height, width

    def _build_anchors_from_shapes(self, anchor_shapes, size):
        num_anchors = len(anchor_shapes)
        anchors = np.zeros((1, 4, num_anchors, size[0], size[1]), dtype=np.float32)
        for anchor_idx in range(num_anchors):
            for y in range(size[0]):
                for x in range(size[1]):
                    anchors[0, 0, anchor_idx, y, x] = (x + 0.5) * self.feature_map_scale
                    anchors[0, 1, anchor_idx, y, x] = (y + 0.5) * self.feature_map_scale
                    anchors[0, 2:, anchor_idx, y, x] = anchor_shapes[anchor_idx]
        return anchors

    @RunOnlyOnce
    def build(self, features):
        anchor_shapes = self._build_anchor_shapes(features)
        size = self._get_anchor_grid_size(features)
        anchors = self._build_anchors_from_shapes(anchor_shapes, size)
        self.register_buffer("anchors", torch.from_numpy(anchors).to(features.device))

    def forward(self, features):
        """
        Create the anchor grid as a tensor.

        :param features: The featuremap on which to create the anchor grid.
        :returns: A tensor representing the anchor grid of shape (1, 4, num_anchor_shapes, h_feat, w_feat).
        """
        self.build(features)
        _,_,h_feat, w_feat = features.shape
        return self.anchors[:, :, :, :h_feat, :w_feat].detach().contiguous()



"""doc
# leanai.model.layers.pooling

> Pooling operations.
"""
import torch
from torch.nn import Module
from torch.nn.functional import max_pool1d as _MaxPooling1D
from torch.nn.functional import max_pool2d as _MaxPooling2D
from torch.nn.functional import avg_pool1d as _AveragePooling1D
from torch.nn.functional import avg_pool2d as _AveragePooling2D
from torch.nn.modules.pooling import AdaptiveAvgPool2d as _AdaptiveAvgPool2d
from leanai.model.layers.flatten import Flatten
from leanai.model.module_registry import add_module


@add_module()
class MaxPooling1D(Module):
    def __init__(self, pool_size=2, stride=None):
        """
        A N max pooling layer.
    
        Computes the max of a N region with stride S.
        This divides the feature map size by S.

        :param pool_size: Size of the region over which is pooled.
        :param stride: The stride defines how the top left corner of the pooling moves across the image. If None then it is same to pool_size resulting in zero overlap between pooled regions.
        """
        super().__init__()
        self.pool_size = pool_size
        self.stride = stride
        if self.stride is None:
            self.stride = self.pool_size
        
    def forward(self, features):
        return _MaxPooling1D(features, self.pool_size, stride=self.stride)


@add_module()
class MaxPooling2D(Module):
    def __init__(self, pool_size=(2, 2), stride=None):
        """
        A NxN max pooling layer.
    
        Computes the max of a NxN region with stride S.
        This divides the feature map size by S.

        :param pool_size: Size of the region over which is pooled.
        :param stride: The stride defines how the top left corner of the pooling moves across the image. If None then it is same to pool_size resulting in zero overlap between pooled regions.
        """
        super().__init__()
        self.pool_size = pool_size
        self.stride = stride
        if self.stride is None:
            self.stride = self.pool_size
        
    def forward(self, features):
        return _MaxPooling2D(features, self.pool_size, stride=self.stride)


@add_module()
class GlobalAveragePooling1D(Module):
    def __init__(self):
        """
        A global average pooling layer.
    
        This computes the global average in N dimension (B, N, C), so that the result is of shape (B, C).
        """
        super().__init__()
        self.flatten = Flatten()
        
    def forward(self, features):
        return self.flatten(_AveragePooling1D(features, features.size()[2:]))


@add_module()
class GlobalAveragePooling2D(Module):
    def __init__(self):
        """
        A global average pooling layer.
    
        This computes the global average in W, H dimension, so that the result is of shape (B, C).
        """
        super().__init__()
        self.flatten = Flatten()

    def forward(self, features):
        return self.flatten(_AveragePooling2D(features, features.size()[2:]))


@add_module()
class AdaptiveAvgPool2D(Module):
    def __init__(self, output_size):
        """
        Wraps AdaptiveAvgPool2d from pytorch
        """
        super().__init__()
        self.pool = _AdaptiveAvgPool2d(output_size)

    def forward(self, features):
        return self.pool(features)



from leanai.data.data_promise import DataPromise, DataPromiseFromBytes, DataPromiseFromFile
from leanai.data.dataset import IIterableDataset, ISequenceDataset, IterableDataset, SequenceDataset
from leanai.data.file_provider import FileProviderIterable, FileProviderSequence
from leanai.data.parser import IParser, Parser
from leanai.data.serializer import ISerializer
from leanai.data.transformer import Transformer


__all__ = [
    "DataPromise", "DataPromiseFromBytes", "DataPromiseFromFile",
    "IIterableDataset", "ISequenceDataset", "IterableDataset", "SequenceDataset",
    "FileProviderIterable", "FileProviderSequence",
    "IParser", "Parser",
    "ISerializer",
    "Transformer",
]



"""doc
# leanai.model.layers.tensor_combiners

> Ways of combining tensors.
"""
import torch
from torch.nn import Module
from leanai.model.module_registry import add_module


@add_module()
class Stack(Module):
    def __init__(self, axis):
        """
        Stack layers along an axis.

        Creates a callable object with the following signature:
        * **tensor_list**: (List[Tensor]) The tensors that should be stacked. A list of length S containing Tensors.
        * **return**: A tensor of shape [..., S, ...] where the position at which S is in the shape is equal to the axis.

        Parameters of the constructor.
        :param axis: (int) The axis along which the stacking happens.
        """
        super().__init__()
        self.axis = axis
        
    def forward(self, *tensor_list):
        return torch.stack(tensor_list, dim=self.axis)


@add_module()
class Concat(Module):
    def __init__(self, axis):
        """
        Concatenate layers along an axis.

        Creates a callable object with the following signature:
        * **tensor_list**: (List[Tensor]) The tensors that should be stacked. A list of length S containing Tensors.
        * **return**: A tensor of shape [..., S * inp_tensor.shape[axis], ...] where the position at which S is in the shape is equal to the axis.

        Parameters of the constructor.
        :param axis: (int) The axis along which the concatenation happens.
        """
        super().__init__()
        self.axis = axis
        
    def forward(self, *tensor_list):
        return torch.cat(tensor_list, dim=self.axis)



"""doc
# leanai.training.losses.loss

> An implementation of the multiloss.
"""
from typing import Union
from torch.nn import Module
from leanai.core.experiment import Experiment


class Loss(Module):
    def __init__(self, parent: Union['Loss', Experiment]):
        """
        Create a loss.

        :param parent: The parent is required for logging.
        """
        super().__init__()
        # put in array so it is invisible to pytorch,
        # otherwise pytorch functions have infinite recursion
        self.__parent = [parent]

    def log(self, name, value, **kwargs):
        """
        Log a value to tensorboard.
        """
        self.__parent[0].log(name, value, **kwargs)



"""doc
# leanai.core.definitions

> Common definitions that are required for your ai development.

Datasplits
* `SPLIT_TRAIN`: Constant used to identify the training datasplit.
* `SPLIT_VAL`: Constant used to identify the validation datasplit. This happens at the end of every epoch and during development.
* `SPLIT_TEST`: Constant used to identify the test datasplit. Testing should only be done once. You should only test the best model that was identified via validation.

Phase of the process
* `PHASE_TRAIN`: Constant specifying the phase to be training of the model.
* `PHASE_VAL`: Constant specifying the phase to be validation of the model. This happens at the end of every epoch and during development.
* `PHASE_TEST`: Constant specifying the phase to be testing of the model. Testing should only be done once. You should only test the best model that was identified via validation.
"""
SPLIT_TRAIN = "train"
SPLIT_VAL = "val"
SPLIT_TEST = "test"

PHASE_TRAIN = "train"
PHASE_VAL = "val"
PHASE_TEST = "test"



"""doc
# leanai.data.file_provider

> The interfaces for providing files.
"""
from typing import Dict, Iterable, Sequence, Iterator
import random

from .data_promise import DataPromise


class FileProviderIterable(Iterable):
    """
    Provides file promises as an iterator.

    The next method returns Dict[str, DataPromise] which is a sample.
    Also implements iter and can optionally implement len.

    A subclass must implement `__next__`.
    """
    def __next__(self) -> Dict[str, DataPromise]:
        raise NotImplementedError("Must be implemented by subclass.")

    def __iter__(self) -> Iterator[Dict[str, DataPromise]]:
        return self

    def __len__(self) -> int:
        raise NotImplementedError("IterableDataset does not always have a length. A subclass does not need to implement this.")


class FileProviderSequence(FileProviderIterable, Sequence):
    def __init__(self, shuffle=True) -> None:
        """
        Provides file promises as an sequence.

        The getitem and next method return Dict[str, DataPromise] which is a sample.
        Also implements iter and len.

        A subclass must implement `__getitem__` and `__len__`.
        """
        super().__init__()
        self.shuffle = shuffle

    def __next__(self) -> Dict[str, DataPromise]:
        if not hasattr(self, "_sample_queue"):
            raise RuntimeError("You must first call iter(...) before you can use next(...).")
        if len(self._sample_queue) == 0:
            delattr(self, "_sample_queue")
            raise StopIteration()
        else:
            return self[self._sample_queue.pop()]

    def __iter__(self) -> Iterator[Dict[str, DataPromise]]:
        self._sample_queue = list(range(len(self)))
        if self.shuffle:
            random.shuffle(self._sample_queue)
        return self

    def __len__(self) -> int:
        raise NotImplementedError("Must be implemented by subclass.")

    def __getitem__(self, idx) -> Dict[str, DataPromise]:
        raise NotImplementedError("Must be implemented by subclass.")



from leanai.data.datasets.fashion_mnist_dataset import FashionMNISTDataset
from leanai.data.datasets.casia_faces_dataset import CasiaFacesDataset
from leanai.data.datasets.coco_dataset import COCODataset
from leanai.data.datasets.images_in_classfolders_dataset import ImagesInClassfoldersDataset



"""doc
# leanai.data.datasets.casia_faces_dataset

> An implementation of a dataset for the casia webfaces.
"""
import numpy as np
from leanai.data.datasets.images_in_classfolders_dataset import ImagesInClassfoldersDataset
from leanai.core.definitions import SPLIT_TRAIN, SPLIT_VAL, SPLIT_TEST


class CasiaFacesDataset(ImagesInClassfoldersDataset):
    def __init__(self, split) -> None:
        super().__init__(split)

    def get_class_id(self, sample_token):
        class_id = super().get_class_id(sample_token)
        return np.array([self.class_mapping[class_id]], dtype=np.uint32)


def test_visualization(data_path):
    import matplotlib.pyplot as plt
    dataset = CasiaFacesDataset(SPLIT_TRAIN)
    image, class_id = dataset[0]
    plt.title(class_id.class_id)
    plt.imshow(image[0])
    plt.show()


if __name__ == "__main__":
    import sys
    test_visualization(sys.argv[1])



"""doc
# leanai.model.layers.dense

> A simple fully connected layer (aka Linear Layer or Dense).
"""
import torch
from torch.nn import Module
from leanai.core.annotations import RunOnlyOnce
from leanai.model.module_registry import add_module

@add_module("FullyConnected")
@add_module()
class Dense(Module):
    def __init__(self, out_features: int, activation=None):
        """
        A simple fully connected layer (aka Linear Layer or Dense).

        It computes Wx+b with optional activation funciton.

        :param out_features: The number of output features.
        :param activation: The activation function that should be added after the fc layer.
        """
        super().__init__()
        self.out_features = out_features
        self.activation = activation

    @RunOnlyOnce
    def build(self, features):
        in_features = features.shape[-1]
        self.linear = torch.nn.Linear(in_features, self.out_features)
        if torch.cuda.is_available():
            self.linear = self.linear.to(torch.device(features.device))

    def forward(self, features):
        self.build(features)

        result = self.linear(features)
        if self.activation is not None:
            result = self.activation(result)
        return result
